# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


def test_loader(
    mocker, swh_config, swh_scheduler_celery_app, swh_scheduler_celery_worker
):
    mock_loader = mocker.patch("swh.loader.mercurial.from_disk.HgLoaderFromDisk.load")
    mock_loader.return_value = {"status": "eventful"}

    res = swh_scheduler_celery_app.send_task(
        "swh.loader.mercurial.tasks.LoadMercurial",
        kwargs={"url": "origin_url", "directory": "/some/repo", "visit_date": "now",},
    )

    assert res
    res.wait()
    assert res.successful()

    assert res.result == {"status": "eventful"}
    mock_loader.assert_called_once_with()


def test_archive_loader(
    mocker, swh_config, swh_scheduler_celery_app, swh_scheduler_celery_worker
):
    mock_loader = mocker.patch(
        "swh.loader.mercurial.from_disk.HgArchiveLoaderFromDisk.load"
    )
    mock_loader.return_value = {"status": "uneventful"}

    res = swh_scheduler_celery_app.send_task(
        "swh.loader.mercurial.tasks.LoadArchiveMercurial",
        kwargs={
            "url": "another_url",
            "archive_path": "/some/tar.tgz",
            "visit_date": "now",
        },
    )
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {"status": "uneventful"}
    mock_loader.assert_called_once_with()



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.loader.mercurial import converters


class TestParseAuthorConverters(unittest.TestCase):
    def test_parse_author_no_email(self):
        self.assertIsNone(converters.parse_author(None))

    def test_parse_author_no_bracket(self):
        actual_author = converters.parse_author(b"someone")

        self.assertEqual(
            actual_author, {"name": None, "email": None, "fullname": b"someone"}
        )

    def test_parse_author_2(self):
        actual_author = converters.parse_author(b"something <wicked")

        self.assertEqual(
            actual_author,
            {"name": b"something", "email": None, "fullname": b"something <wicked"},
        )

    def test_parse_author_3(self):
        actual_author = converters.parse_author(b"something >wicked")

        self.assertEqual(
            actual_author,
            {"name": None, "email": None, "fullname": b"something >wicked"},
        )

    def test_parse_author_4(self):
        actual_author = converters.parse_author(b"something <")

        self.assertEqual(
            actual_author,
            {"name": b"something", "email": None, "fullname": b"something <"},
        )

    def test_parse_author_5(self):
        actual_author = converters.parse_author(b"<only>")

        self.assertEqual(
            actual_author, {"name": None, "email": b"only", "fullname": b"<only>"}
        )

    def test_parse_author_6(self):
        actual_author = converters.parse_author(b"  <something>")

        self.assertEqual(
            actual_author,
            {"name": b" ", "email": b"something", "fullname": b"  <something>"},
        )

    def test_parse_author_normal(self):
        actual_author = converters.parse_author(b"someone <awesome>")

        self.assertEqual(
            actual_author,
            {"name": b"someone", "email": b"awesome", "fullname": b"someone <awesome>"},
        )



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from codecs import escape_decode  # type: ignore
import json
from pathlib import Path
import re
import subprocess
from typing import Any, Dict, Iterator, List, NamedTuple, Optional, Union

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import click

from swh.loader.mercurial.utils import get_minimum_env
from swh.model.cli import identify_object
from swh.model.hashutil import hash_to_bytehex
from swh.model.identifiers import CoreSWHID, ObjectType, normalize_timestamp
from swh.model.model import RevisionType

TAG_PATTERN = re.compile(b"([0-9A-Fa-f]{40}) +(.+)")


class HgAuthor(NamedTuple):
    """Represent a Mercurial revision author."""

    fullname: bytes
    """full name of the author"""

    name: Optional[bytes]
    """name of the author"""

    email: Optional[bytes]
    """email of the author"""

    @staticmethod
    def from_bytes(data: bytes) -> "HgAuthor":
        """Convert bytes to an HgAuthor named tuple.

        Expected format: "name <email>"
        """
        from swh.loader.mercurial.converters import parse_author

        result = parse_author(data)
        return HgAuthor(
            fullname=result["fullname"], name=result["name"], email=result["email"]
        )

    def to_dict(self) -> Dict[str, Optional[bytes]]:
        return {"fullname": self.fullname, "name": self.name, "email": self.email}


HG_REVISION_TEMPLATE = "\n".join(
    [
        "node_id:{node}",
        "author:{author}",
        "timestamp_offset:{date|json}",
        "p1:{p1.node}",
        "p2:{p2.node}",
        "extras:{join(extras, '\nextras:')}",
    ]
)  # Log template for HgRevision.from_bytes

NULL_NODE_ID = b"0" * 40  # Value used when no parent


class HgRevision(NamedTuple):
    """Represent a Mercurial revision."""

    node_id: bytes
    """raw bytes of the revision hash"""

    author: HgAuthor
    """author of the revision"""

    timestamp: bytes
    """timestamp of the revision"""

    offset: bytes
    """offset of the revision"""

    parents: List[bytes]
    """hex bytes of the revision's parents"""

    extras: Dict[bytes, bytes]
    """metadata of the revision"""

    description: bytes
    """description of the revision"""

    @staticmethod
    def from_bytes(data: bytes, description: bytes) -> "HgRevision":
        """Convert bytes to an HgRevision named tuple.

        Expected data format:
        '''
        node_id:{node}
        author:{author}
        timestamp_offset:[{timestamp}, {offset}]
        p1:{p1}
        p2:{p2}
        extras:{key1}={value1}
        ...
        extras:{keyn}={value}
        '''

        """
        lines = data.split(b"\n")
        tuples = [line.split(b":", 1) for line in lines]
        fields: Dict[str, Any] = {
            "parents": [],
            "extras": {},
            "description": description,
        }
        for key, value in tuples:
            if key == b"timestamp_offset":
                timestamp, offset = json.loads(value)
                fields["timestamp"] = timestamp
                fields["offset"] = offset
            elif key in (b"p1", b"p2"):
                if value != NULL_NODE_ID:
                    fields["parents"].append(value)
            elif key == b"extras":
                extra_key, extra_value = value.split(b"=", 1)
                fields["extras"][extra_key] = extra_value
            elif key == b"author":
                fields["author"] = HgAuthor.from_bytes(value)
            else:
                fields[key.decode()] = value

        return HgRevision(**fields)

    def branch(self) -> bytes:
        return self.extras.get(b"branch", b"default")

    def to_dict(self) -> Dict:
        """Convert a HgRevision to a dict for SWHID computation"""
        date = normalize_timestamp(int(self.timestamp))

        extra_headers = [
            (b"time_offset_seconds", str(self.offset).encode("utf-8")),
        ]

        for key, value in self.extras.items():
            if key == b"branch" and value == b"default":
                # branch default is skipped to match historical implementation
                continue
            if key == b"transplant_source":
                # transplant_source is converted to hex
                # to match historical implementation
                value = hash_to_bytehex(escape_decode(value)[0])
            extra_headers.append((key, value))

        author = self.author.to_dict()

        return {
            "author": author,
            "date": date,
            "committer": author,
            "committer_date": date,
            "type": RevisionType.MERCURIAL.value,
            "message": self.description,
            "metadata": {"node": self.node_id},
            "extra_headers": tuple(extra_headers),
            "synthetic": False,
            "parents": self.parents,
        }


class HgBranch(NamedTuple):
    """Represent a Mercurial branch."""

    name: bytes
    """name of the branch"""

    node_id: bytes
    """row bytes of the target revision hash"""


class HgTag(NamedTuple):
    """Represent a Mercurial tag."""

    name: bytes
    """name of the tag"""

    node_id: bytes
    """hex bytes of the target revision"""


class Hg:
    """Provide methods to extract data from a Mercurial repository."""

    def __init__(self, repository_root: Path) -> None:
        self._root = repository_root

    def _output(self, *args) -> bytes:
        """Return the outpout of a `hg` call."""
        return subprocess.check_output(
            ["hg", *args], cwd=self._root, env=get_minimum_env()
        )

    def _call(self, *args) -> None:
        """Perform a `hg` call."""
        subprocess.check_call(
            ["hg", *args],
            cwd=self._root,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            env=get_minimum_env(),
        )

    def root(self) -> Path:
        """Return the root of the Mercurial repository."""
        return self._root

    def log(self, rev: Optional[Union[bytes, str]] = None) -> List[HgRevision]:
        """Return the specified revisions of the Mercurial repository.

        Mercurial revsets are supported. (See `hg help revsets`)

        If no revision range is specified, return all revisions".
        """
        if rev:
            node_ids = self._output("log", "-r", rev, "-T", "{node}\n").splitlines()
        else:
            node_ids = self._output("log", "-T", "{node}\n").splitlines()

        revisions = [self._revision(node_id) for node_id in reversed(node_ids)]

        return revisions

    def _revision(self, revision: bytes) -> HgRevision:
        data = self._output("log", "-r", revision, "-T", HG_REVISION_TEMPLATE)

        # hg log strips the description so the raw description has to be taken
        # from debugdata
        # The description follows some metadata and is separated from them
        # by an empty line
        _, desc = self._output("debugdata", "-c", revision).split(b"\n\n", 1)

        return HgRevision.from_bytes(data, desc)

    def up(self, rev: bytes) -> None:
        """Update the repository working directory to the specified revision."""
        self._call("up", rev)

    def branches(self) -> List[HgBranch]:
        """List the repository named branches."""
        output = self._output("branches", "-T", "{branch}\n{node}\n\n").strip()

        branches = []

        for block in output.split(b"\n\n"):
            name, node_id = block.splitlines()
            branches.append(HgBranch(name=name, node_id=node_id))

        return branches

    def tip(self) -> HgRevision:
        """Return the `tip` node-id."""
        return self.log("tip")[0]

    def tags(self) -> List[HgTag]:
        """Return the repository's tags as defined in the `.hgtags` file.

        `.hgtags` being like any other repository's tracked file, its content can vary
        from revision to revision. The returned value therefore depends on the current
        revision of the repository.
        """
        hgtags = self._root / ".hgtags"

        tags = {}

        if hgtags.is_file():
            for line in hgtags.read_bytes().splitlines():
                match = TAG_PATTERN.match(line)
                if match is None:
                    continue
                node_id, name = match.groups()
                tags[node_id] = name

        return [HgTag(name=name, node_id=node_id) for node_id, name in tags.items()]


@click.group()
@click.option(
    "--directory",
    "-d",
    help=("Path to the Mercurial repository. If unset, the current directory is used"),
)
@click.pass_context
def main(ctx, directory=None):
    """Compute the Software Heritage persistent identifier (SWHID) for the given
       source code object(s).

    For more details about SWHIDs see:

    https://docs.softwareheritage.org/devel/swh-model/persistent-identifiers.html
    """
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    ctx.ensure_object(dict)

    root = Path(directory) if directory else Path()
    if not root.exists():
        raise IOError(f"{root!r} does not exists")

    ctx.obj["HG_ROOT"] = root


def identify_directory(path: Path) -> CoreSWHID:
    """Return the SWHID of the given path."""
    return CoreSWHID.from_string(
        identify_object(
            "directory", follow_symlinks=True, exclude_patterns=[b".hg"], obj=str(path)
        )
    )


class RevisionIdentity(NamedTuple):
    """Represent a swh revision identity."""

    swhid: CoreSWHID
    """SWH Identifier of the revision."""

    node_id: bytes
    """node_id hex bytes"""

    directory_swhid: CoreSWHID
    """SWH Identifier of the directory"""

    def dir_uri(self) -> str:
        """Return the SWHID uri of the revision's directory."""
        return f"{self.directory_swhid}\t{self.node_id.decode()}"

    def __str__(self) -> str:
        """Return the string representation of a RevisionIdentity."""
        return f"{self.swhid}\t{self.node_id.decode()}"


def identify_revision(
    hg: Hg,
    rev: Optional[bytes] = None,
    node_id_2_swhid: Optional[Dict[bytes, CoreSWHID]] = None,
) -> Iterator[RevisionIdentity]:
    """Return the repository revision identities.

    Args:
        hg: A `Hg` repository instance
        rev: An optional revision or Mercurial revsets (See `hg help revsets`)
             If not provided all the repository revisions will be computed.
        node_id_2_swhid: An optional cache mapping hg node ids to SWHIDs
            It will be updated in place with new mappings.
    """
    from swh.model.model import Revision

    if node_id_2_swhid is None:
        node_id_2_swhid = {}

    for revision in hg.log(rev):
        data = revision.to_dict()

        hg.up(revision.node_id)
        directory_swhid = identify_directory(hg.root())
        data["directory"] = directory_swhid.object_id

        parents = []
        for parent in data["parents"]:
            if parent not in node_id_2_swhid:
                parent_revision = next(identify_revision(hg, parent, node_id_2_swhid))
                node_id_2_swhid[parent] = parent_revision.swhid
            assert node_id_2_swhid[parent].object_type == ObjectType.REVISION
            parents.append(node_id_2_swhid[parent].object_id)
        data["parents"] = parents

        revision_swhid = Revision.from_dict(data).swhid()
        node_id_2_swhid[revision.node_id] = revision_swhid

        yield RevisionIdentity(
            swhid=revision_swhid,
            node_id=revision.node_id,
            directory_swhid=directory_swhid,
        )


class ReleaseIdentity(NamedTuple):
    """Represent a swh release identity."""

    swhid: CoreSWHID
    """SWH Identifier of the release."""

    node_id: bytes
    """node_id hex bytes"""

    name: bytes
    """name of the release"""

    def __str__(self) -> str:
        """Return the string representation of a ReleaseIdentity."""
        return f"{self.swhid}\t{self.name.decode()}"


def identify_release(
    hg: Hg, node_id_2_swhid: Optional[Dict[bytes, CoreSWHID]] = None,
) -> Iterator[ReleaseIdentity]:
    """Return the repository's release identities.

    Args:
        hg: A `Hg` repository instance
        node_id_2_swhid: An optional cache mapping hg node ids to SWHIDs
            If not provided it will be computed using `identify_revision`.
    """
    from swh.model.model import ObjectType as ModelObjectType
    from swh.model.model import Release

    if node_id_2_swhid is None:
        node_id_2_swhid = {
            revision.node_id: revision.swhid for revision in identify_revision(hg)
        }

    for tag in hg.tags():
        assert node_id_2_swhid[tag.node_id].object_type == ObjectType.REVISION
        data = {
            "name": tag.name,
            "target": node_id_2_swhid[tag.node_id].object_id,
            "target_type": ModelObjectType.REVISION.value,
            "message": None,
            "metadata": None,
            "synthetic": False,
            "author": {"name": None, "email": None, "fullname": b""},
            "date": None,
        }

        release_swhid = Release.from_dict(data).swhid()

        yield ReleaseIdentity(
            swhid=release_swhid, node_id=tag.node_id, name=tag.name,
        )


def identify_snapshot(
    hg: Hg,
    node_id_2_swhid: Optional[Dict[bytes, CoreSWHID]] = None,
    releases: Optional[List[ReleaseIdentity]] = None,
) -> CoreSWHID:
    """Return the repository snapshot identity.

    Args:
        hg: A `Hg` repository instance
        node_id_2_swhid: An optional cache mapping hg node ids to SWHIDs
             If not provided it will be computed using `identify_revision`.
        release: an optional list of `ReleaseIdentity`.
            If not provided it will be computed using `identify_release`.
    """
    from swh.model.model import Snapshot, TargetType

    if node_id_2_swhid is None:
        node_id_2_swhid = {
            revision.node_id: revision.swhid for revision in identify_revision(hg)
        }

    if releases is None:
        releases = [release for release in identify_release(hg, node_id_2_swhid)]

    branches = {}

    tip = hg.tip()
    branches[b"HEAD"] = {
        "target": tip.branch(),
        "target_type": TargetType.ALIAS.value,
    }

    for branch in hg.branches():
        assert node_id_2_swhid[branch.node_id].object_type == ObjectType.REVISION
        branches[branch.name] = {
            "target": node_id_2_swhid[branch.node_id].object_id,
            "target_type": TargetType.REVISION.value,
        }

    for release in releases:
        assert release.swhid.object_type == ObjectType.RELEASE
        branches[release.name] = {
            "target": release.swhid.object_id,
            "target_type": TargetType.RELEASE.value,
        }

    return Snapshot.from_dict({"branches": branches}).swhid()


@main.command()
@click.argument("rev", required=False)
@click.pass_context
def revision(ctx, rev):
    """Compute the SWHID of a given revision.

    If specified REV allow to select a single or multiple revisions
    (using the Mercurial revsets language: `hg help revsets`)
    """
    hg = Hg(ctx.obj["HG_ROOT"])

    for identity in identify_revision(hg, rev):
        click.echo(identity)


@main.command()
@click.pass_context
def snapshot(ctx):
    """Compute the SWHID of the snapshot."""
    root = ctx.obj["HG_ROOT"]
    hg = Hg(root)

    snapshot_swhid = identify_snapshot(hg)

    click.echo(f"{snapshot_swhid}\t{root}")


@main.command()
@click.pass_context
def all(ctx):
    """Compute the SWHID of all the repository objects."""
    root = ctx.obj["HG_ROOT"]
    hg = Hg(root)

    dir_uris = []
    rev_uris = []
    rel_uris = []

    node_id_2_swhid = {}
    for revision in identify_revision(hg):
        dir_uris.append(revision.dir_uri())
        rev_uris.append(str(revision))
        node_id_2_swhid[revision.node_id] = revision.swhid

    releases = []
    for release in identify_release(hg, node_id_2_swhid):
        rel_uris.append(str(release))
        releases.append(release)

    snapshot_swhid = identify_snapshot(hg, node_id_2_swhid, releases)

    for uri in dir_uris + rev_uris + rel_uris:
        click.echo(uri)

    click.echo(f"{snapshot_swhid}\t{root}")


if __name__ == "__main__":
    main()



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timezone
import os
from typing import Dict, Optional, Union

from dateutil.parser import parse


def parse_visit_date(visit_date: Optional[Union[datetime, str]]) -> Optional[datetime]:
    """Convert visit date from either None, a string or a datetime to either None or
       datetime.

    """
    if visit_date is None:
        return None

    if isinstance(visit_date, datetime):
        return visit_date

    if visit_date == "now":
        return datetime.now(tz=timezone.utc)

    if isinstance(visit_date, str):
        return parse(visit_date)

    raise ValueError(f"invalid visit date {visit_date!r}")


def get_minimum_env() -> Dict[str, str]:
    """Return the smallest viable environment for `hg` suprocesses"""
    env = {
        "HGPLAIN": "",  # Tells Mercurial to disable output customization
        "HGRCPATH": "",  # Tells Mercurial to ignore user's config files
        "HGRCSKIPREPO": "",  # Tells Mercurial to ignore repo's config file
    }
    path = os.environ.get("PATH")
    if path:
        # Sometimes (in tests for example), there is no PATH. An empty PATH could be
        # interpreted differently than a lack of PATH by some programs.
        env["PATH"] = path
    return env



# Copyright (C) 2019-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Any, Dict

import pytest


@pytest.fixture
def swh_storage_backend_config(swh_storage_backend_config):
    """Basic pg storage configuration with no journal collaborator
    (to avoid pulling optional dependency on clients of this fixture)

    """
    return {
        "cls": "filter",
        "storage": {
            "cls": "buffer",
            "min_batch_size": {
                "content": 10,
                "content_bytes": 100 * 1024 * 1024,
                "directory": 10,
                "revision": 10,
                "release": 10,
            },
            "storage": swh_storage_backend_config,
        },
    }


@pytest.fixture
def swh_loader_config(swh_storage_backend_config, tmp_path) -> Dict[str, Any]:
    return {
        "storage": swh_storage_backend_config,
        "max_content_size": 104857600,
        "temp_directory": str(tmp_path),
    }


@pytest.fixture(autouse=True, scope="session")
def swh_mercurial_set_plain():
    """Mercurial is customizable by users, so we use built-in environment
    variables to turn off all customization in tests.
    """
    os.environ["HGPLAIN"] = ""
    os.environ["HGRCPATH"] = ""
    os.environ["HGRCSKIPREPO"] = ""



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from pathlib import Path
from typing import NamedTuple, Set

from swh.loader.core.loader import BaseLoader
from swh.model.hashutil import hash_to_bytes


class ExpectedSwhids(NamedTuple):
    """List the of swhids expected from the loader."""

    directories: Set[str]
    """Hex swhid of the root directory of each revision."""

    revisions: Set[str]
    """Hex swhid of each revision."""

    releases: Set[str]
    """Hex swhid of each release."""

    snapshot: str
    """Hex swhid of the snapshot."""

    @staticmethod
    def load(path: Path) -> "ExpectedSwhids":
        """Load expected swhids from a json file.

        See `build.py` in the data directory on how to extract that json file
        from an existing repository or archive.
        """
        data = json.load(open(path))
        return ExpectedSwhids(
            directories=set(data["directories"]),
            revisions=set(data["revisions"]),
            releases=set(data["releases"]),
            snapshot=data["snapshot"],
        )


class LoaderChecker:
    """Check the swhids produced by a BaseLoader."""

    def __init__(self, loader: BaseLoader, expected: ExpectedSwhids) -> None:
        self._loader = loader
        self._expected = expected

    def check(self) -> None:
        """Check loader's outputs."""
        assert self._loader.load() == {"status": "eventful"}

        missing_directories = self._loader.storage.directory_missing(
            [hash_to_bytes(id) for id in self._expected.directories]
        )
        assert list(missing_directories) == []

        missing_revisions = self._loader.storage.revision_missing(
            [hash_to_bytes(id) for id in self._expected.revisions]
        )
        assert list(missing_revisions) == []

        missing_releases = self._loader.storage.release_missing(
            [hash_to_bytes(id) for id in self._expected.releases]
        )
        assert list(missing_releases) == []

        snapshot = self._loader.storage.snapshot_get(
            hash_to_bytes(self._expected.snapshot)
        )
        assert snapshot is not None



# Copyright (C) 2017-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""This document contains a SWH loader for ingesting repository data
from Mercurial version 2 bundle files.

"""

# NOTE: The code here does expensive work twice in places because of the
# intermediate need to check for what is missing before sending to the database
# and the desire to not juggle very large amounts of data.

# TODO: Decide whether to also serialize to disk and read back more quickly
#       from there. Maybe only for very large repos and fast drives.
# - Avi


import datetime
import os
from queue import Empty
import random
import re
from shutil import rmtree
from tempfile import mkdtemp
import time
from typing import Any, Dict, Iterable, List, Optional

import billiard
import hglib
from hglib.error import CommandError

from swh.loader.core.loader import DVCSLoader
from swh.loader.core.utils import clean_dangling_folders
from swh.loader.exception import NotFound
from swh.loader.mercurial.utils import get_minimum_env
from swh.model import identifiers
from swh.model.hashutil import (
    DEFAULT_ALGORITHMS,
    MultiHash,
    hash_to_bytehex,
    hash_to_bytes,
)
from swh.model.model import (
    BaseContent,
    Content,
    Directory,
    ExtID,
    ObjectType,
    Origin,
    Person,
    Release,
    Revision,
    RevisionType,
    Sha1Git,
    SkippedContent,
    Snapshot,
    SnapshotBranch,
    TargetType,
    TimestampWithTimezone,
)
from swh.storage.algos.origin import origin_get_latest_visit_status
from swh.storage.interface import StorageInterface

from . import converters
from .archive_extract import tmp_extract
from .bundle20_reader import Bundle20Reader
from .converters import PRIMARY_ALGO as ALGO
from .objects import SelectiveCache, SimpleTree

TAG_PATTERN = re.compile("[0-9A-Fa-f]{40}")

TEMPORARY_DIR_PREFIX_PATTERN = "swh.loader.mercurial."

HEAD_POINTER_NAME = b"tip"

EXTID_TYPE = "hg-nodeid"


class CommandErrorWrapper(Exception):
    """This exception is raised in place of a 'CommandError'
       exception (raised by the underlying hglib library)

       This is needed because billiard.Queue is serializing the
       queued object and as CommandError doesn't have a constructor without
       parameters, the deserialization is failing
    """

    def __init__(self, err: Optional[bytes]):
        self.err = err


class CloneTimeoutError(Exception):
    pass


class HgBundle20Loader(DVCSLoader):
    """Mercurial loader able to deal with remote or local repository.

    """

    visit_type = "hg"

    def __init__(
        self,
        storage: StorageInterface,
        url: str,
        visit_date: Optional[datetime.datetime] = None,
        directory: Optional[str] = None,
        logging_class="swh.loader.mercurial.Bundle20Loader",
        bundle_filename: Optional[str] = "HG20_none_bundle",
        reduce_effort: bool = False,
        temp_directory: str = "/tmp",
        cache1_size: int = 800 * 1024 * 1024,
        cache2_size: int = 800 * 1024 * 1024,
        clone_timeout_seconds: int = 7200,
        save_data_path: Optional[str] = None,
        max_content_size: Optional[int] = None,
    ):
        super().__init__(
            storage=storage,
            logging_class=logging_class,
            save_data_path=save_data_path,
            max_content_size=max_content_size,
        )
        self.origin_url = url
        self.visit_date = visit_date
        self.directory = directory
        self.bundle_filename = bundle_filename
        self.reduce_effort_flag = reduce_effort
        self.empty_repository = None
        self.temp_directory = temp_directory
        self.cache1_size = cache1_size
        self.cache2_size = cache2_size
        self.clone_timeout = clone_timeout_seconds
        self.working_directory = None
        self.bundle_path = None
        self.heads: Dict[bytes, Any] = {}
        self.releases: Dict[bytes, Any] = {}
        self.last_snapshot_id: Optional[bytes] = None
        self.old_environ = os.environ.copy()
        os.environ.clear()
        os.environ.update(get_minimum_env())

    def pre_cleanup(self):
        """Cleanup potential dangling files from prior runs (e.g. OOM killed
           tasks)

        """
        clean_dangling_folders(
            self.temp_directory,
            pattern_check=TEMPORARY_DIR_PREFIX_PATTERN,
            log=self.log,
        )

    def cleanup(self):
        """Clean temporary working directory

        """
        os.environ.clear()
        os.environ.update(self.old_environ)
        if self.bundle_path and os.path.exists(self.bundle_path):
            self.log.debug("Cleanup up working bundle %s" % self.bundle_path)
            os.unlink(self.bundle_path)
        if self.working_directory and os.path.exists(self.working_directory):
            self.log.debug(
                "Cleanup up working directory %s" % (self.working_directory,)
            )
            rmtree(self.working_directory)

    def get_heads(self, repo):
        """Read the closed branches heads (branch, bookmarks) and returns a
           dict with key the branch_name (bytes) and values the tuple
           (pointer nature (bytes), mercurial's node id
           (bytes)). Those needs conversion to swh-ids. This is taken
           care of in get_revisions.

        """
        b = {}
        for _, node_hash_id, pointer_nature, branch_name, *_ in repo.heads():
            b[branch_name] = (pointer_nature, hash_to_bytes(node_hash_id.decode()))

        bookmarks = repo.bookmarks()
        if bookmarks and bookmarks[0]:
            for bookmark_name, _, target_short in bookmarks[0]:
                target = repo[target_short].node()
                b[bookmark_name] = (None, hash_to_bytes(target.decode()))

        return b

    def prepare_origin_visit(self) -> None:
        self.origin = Origin(url=self.origin_url)
        visit_status = origin_get_latest_visit_status(
            self.storage, self.origin_url, require_snapshot=True
        )
        self.last_snapshot_id = None if visit_status is None else visit_status.snapshot

    @staticmethod
    def clone_with_timeout(log, origin, destination, timeout):
        queue = billiard.Queue()
        start = time.monotonic()

        def do_clone(queue, origin, destination):
            try:
                result = hglib.clone(source=origin, dest=destination, noupdate=True)
            except CommandError as e:
                # the queued object need an empty constructor to be deserialized later
                queue.put(CommandErrorWrapper(e.err))
            except BaseException as e:
                queue.put(e)
            else:
                queue.put(result)

        process = billiard.Process(target=do_clone, args=(queue, origin, destination))
        process.start()

        while True:
            try:
                result = queue.get(timeout=0.1)
                break
            except Empty:
                duration = time.monotonic() - start
                if timeout and duration > timeout:
                    log.warning(
                        "Timeout cloning `%s` within %s seconds", origin, timeout
                    )
                    process.terminate()
                    process.join()
                    raise CloneTimeoutError(origin, timeout)
                continue

        process.join()

        if isinstance(result, Exception):
            raise result from None

        return result

    def prepare(self):
        """Prepare the necessary steps to load an actual remote or local
           repository.

           To load a local repository, pass the optional directory
           parameter as filled with a path to a real local folder.

           To load a remote repository, pass the optional directory
           parameter as None.

        Args:
            origin_url (str): Origin url to load
            visit_date (str/datetime): Date of the visit
            directory (str/None): The local directory to load

        """
        self.branches = {}
        self.tags = []
        self.releases = {}
        self.node_2_rev = {}
        self.heads = {}
        self.extids = []

        directory = self.directory

        if not directory:  # remote repository
            self.working_directory = mkdtemp(
                prefix=TEMPORARY_DIR_PREFIX_PATTERN,
                suffix="-%s" % os.getpid(),
                dir=self.temp_directory,
            )
            os.makedirs(self.working_directory, exist_ok=True)
            self.hgdir = self.working_directory

            self.log.debug(
                "Cloning %s to %s with timeout %s seconds",
                self.origin_url,
                self.hgdir,
                self.clone_timeout,
            )

            try:
                self.clone_with_timeout(
                    self.log, self.origin_url, self.hgdir, self.clone_timeout
                )
            except CommandErrorWrapper as e:
                for msg in [
                    b"does not appear to be an hg repository",
                    b"404: not found",
                    b"name or service not known",
                ]:
                    if msg in e.err.lower():
                        raise NotFound(e.args[0]) from None
                raise e

        else:  # local repository
            self.working_directory = None
            self.hgdir = directory

        self.bundle_path = os.path.join(self.hgdir, self.bundle_filename)
        self.log.debug("Bundling at %s" % self.bundle_path)

        with hglib.open(self.hgdir) as repo:
            self.heads = self.get_heads(repo)
            repo.bundle(bytes(self.bundle_path, "utf-8"), all=True, type=b"none-v2")

        self.cache_filename1 = os.path.join(
            self.hgdir, "swh-cache-1-%s" % (hex(random.randint(0, 0xFFFFFF))[2:],)
        )
        self.cache_filename2 = os.path.join(
            self.hgdir, "swh-cache-2-%s" % (hex(random.randint(0, 0xFFFFFF))[2:],)
        )

        try:
            self.br = Bundle20Reader(
                bundlefile=self.bundle_path,
                cache_filename=self.cache_filename1,
                cache_size=self.cache1_size,
            )
        except FileNotFoundError:
            # Empty repository! Still a successful visit targeting an
            # empty snapshot
            self.log.warn("%s is an empty repository!" % self.hgdir)
            self.empty_repository = True
        else:
            self.reduce_effort = set()
            if self.reduce_effort_flag:
                now = datetime.datetime.now(tz=datetime.timezone.utc)
                if (now - self.visit_date).days > 1:
                    # Assuming that self.visit_date would be today for
                    # a new visit, treat older visit dates as
                    # indication of wanting to skip some processing
                    # effort.
                    for header, commit in self.br.yield_all_changesets():
                        ts = commit["time"].timestamp()
                        if ts < self.visit_date.timestamp():
                            self.reduce_effort.add(header["node"])

    def has_contents(self):
        return not self.empty_repository

    def has_directories(self):
        return not self.empty_repository

    def has_revisions(self):
        return not self.empty_repository

    def has_releases(self):
        return not self.empty_repository

    def fetch_data(self):
        """Fetch the data from the data source."""
        pass

    def get_contents(self) -> Iterable[BaseContent]:
        """Get the contents that need to be loaded."""

        # NOTE: This method generates blobs twice to reduce memory usage
        # without generating disk writes.
        self.file_node_to_hash = {}
        hash_to_info = {}
        self.num_contents = 0
        contents: Dict[bytes, BaseContent] = {}
        missing_contents = set()

        for blob, node_info in self.br.yield_all_blobs():
            self.num_contents += 1
            file_name = node_info[0]
            header = node_info[2]

            length = len(blob)
            if header["linknode"] in self.reduce_effort:
                algorithms = set([ALGO])
            else:
                algorithms = DEFAULT_ALGORITHMS
            h = MultiHash.from_data(blob, hash_names=algorithms)
            content = h.digest()
            content["length"] = length
            blob_hash = content[ALGO]
            self.file_node_to_hash[header["node"]] = blob_hash

            if header["linknode"] in self.reduce_effort:
                continue

            hash_to_info[blob_hash] = node_info
            if self.max_content_size is not None and length >= self.max_content_size:
                contents[blob_hash] = SkippedContent(
                    status="absent", reason="Content too large", **content
                )
            else:
                contents[blob_hash] = Content(data=blob, status="visible", **content)

            if file_name == b".hgtags":
                # https://www.mercurial-scm.org/wiki/GitConcepts#Tag_model
                # overwrite until the last one
                self.tags = (t for t in blob.split(b"\n") if t != b"")

        if contents:
            missing_contents = set(
                self.storage.content_missing(
                    [c.to_dict() for c in contents.values()], key_hash=ALGO
                )
            )

        # Clusters needed blobs by file offset and then only fetches the
        # groups at the needed offsets.
        focs: Dict[int, Dict[bytes, bytes]] = {}  # "file/offset/contents"
        for blob_hash in missing_contents:
            _, file_offset, header = hash_to_info[blob_hash]
            focs.setdefault(file_offset, {})
            focs[file_offset][header["node"]] = blob_hash

        for offset, node_hashes in sorted(focs.items()):
            for header, data, *_ in self.br.yield_group_objects(group_offset=offset):
                node = header["node"]
                if node in node_hashes:
                    blob, meta = self.br.extract_meta_from_blob(data)
                    content = contents.pop(node_hashes[node], None)
                    if content:
                        if (
                            self.max_content_size is not None
                            and len(blob) >= self.max_content_size
                        ):
                            yield SkippedContent.from_data(
                                blob, reason="Content too large"
                            )
                        else:
                            yield Content.from_data(blob)

    def load_directories(self):
        """This is where the work is done to convert manifest deltas from the
        repository bundle into SWH directories.

        """
        self.mnode_to_tree_id = {}
        cache_hints = self.br.build_manifest_hints()

        def tree_size(t):
            return t.size()

        self.trees = SelectiveCache(
            cache_hints=cache_hints,
            size_function=tree_size,
            filename=self.cache_filename2,
            max_size=self.cache2_size,
        )

        tree = SimpleTree()
        for header, added, removed in self.br.yield_all_manifest_deltas(cache_hints):
            node = header["node"]
            basenode = header["basenode"]
            tree = self.trees.fetch(basenode) or tree  # working tree

            for path in removed.keys():
                tree = tree.remove_tree_node_for_path(path)
            for path, info in added.items():
                file_node, is_symlink, perms_code = info
                tree = tree.add_blob(
                    path, self.file_node_to_hash[file_node], is_symlink, perms_code
                )

            if header["linknode"] in self.reduce_effort:
                self.trees.store(node, tree)
            else:
                new_dirs = []
                self.mnode_to_tree_id[node] = tree.hash_changed(new_dirs)
                self.trees.store(node, tree)
                yield header, tree, new_dirs

    def get_directories(self) -> Iterable[Directory]:
        """Compute directories to load

        """
        dirs: Dict[Sha1Git, Directory] = {}
        self.num_directories = 0
        for _, _, new_dirs in self.load_directories():
            for d in new_dirs:
                self.num_directories += 1
                dirs[d["id"]] = Directory.from_dict(d)

        missing_dirs: List[Sha1Git] = list(dirs.keys())
        if missing_dirs:
            missing_dirs = list(self.storage.directory_missing(missing_dirs))

        for _id in missing_dirs:
            yield dirs[_id]

    def get_revisions(self) -> Iterable[Revision]:
        """Compute revisions to load

        """
        revisions = {}
        self.num_revisions = 0
        for header, commit in self.br.yield_all_changesets():
            if header["node"] in self.reduce_effort:
                continue

            self.num_revisions += 1
            date_dict = identifiers.normalize_timestamp(int(commit["time"].timestamp()))
            author_dict = converters.parse_author(commit["user"])
            if commit["manifest"] == Bundle20Reader.NAUGHT_NODE:
                directory_id = SimpleTree().hash_changed()
            else:
                directory_id = self.mnode_to_tree_id[commit["manifest"]]

            extra_headers = [
                (
                    b"time_offset_seconds",
                    str(commit["time_offset_seconds"]).encode("utf-8"),
                )
            ]
            extra = commit.get("extra")
            if extra:
                for e in extra.split(b"\x00"):
                    k, v = e.split(b":", 1)
                    # transplant_source stores binary reference to a changeset
                    # prefer to dump hexadecimal one in the revision metadata
                    if k == b"transplant_source":
                        v = hash_to_bytehex(v)
                    extra_headers.append((k, v))

            parents = []
            p1 = self.node_2_rev.get(header["p1"])
            p2 = self.node_2_rev.get(header["p2"])
            if p1:
                parents.append(p1)
            if p2:
                parents.append(p2)

            revision = Revision(
                author=Person.from_dict(author_dict),
                date=TimestampWithTimezone.from_dict(date_dict),
                committer=Person.from_dict(author_dict),
                committer_date=TimestampWithTimezone.from_dict(date_dict),
                type=RevisionType.MERCURIAL,
                directory=directory_id,
                message=commit["message"],
                extra_headers=tuple(extra_headers),
                synthetic=False,
                parents=tuple(parents),
            )

            self.node_2_rev[header["node"]] = revision.id
            revisions[revision.id] = revision

            revision_swhid = identifiers.CoreSWHID(
                object_type=identifiers.ObjectType.REVISION, object_id=revision.id,
            )
            self.extids.append(
                ExtID(
                    extid_type=EXTID_TYPE, extid=header["node"], target=revision_swhid
                )
            )

        # Converts heads to use swh ids
        self.heads = {
            branch_name: (pointer_nature, self.node_2_rev[node_id])
            for branch_name, (pointer_nature, node_id) in self.heads.items()
        }

        missing_revs = set(revisions.keys())
        if missing_revs:
            missing_revs = set(self.storage.revision_missing(list(missing_revs)))

        for rev in missing_revs:
            yield revisions[rev]
        self.mnode_to_tree_id = None

    def _read_tag(self, tag, split_byte=b" "):
        node, *name = tag.split(split_byte)
        name = split_byte.join(name)
        return node, name

    def get_releases(self) -> Iterable[Release]:
        """Get the releases that need to be loaded."""
        self.num_releases = 0
        releases = {}
        missing_releases = set()
        for t in self.tags:
            self.num_releases += 1
            node, name = self._read_tag(t)
            node = node.decode()
            node_bytes = hash_to_bytes(node)
            if not TAG_PATTERN.match(node):
                self.log.warn("Wrong pattern (%s) found in tags. Skipping" % (node,))
                continue
            if node_bytes not in self.node_2_rev:
                self.log.warn(
                    "No matching revision for tag %s "
                    "(hg changeset: %s). Skipping" % (name.decode(), node)
                )
                continue
            tgt_rev = self.node_2_rev[node_bytes]
            release = Release(
                name=name,
                target=tgt_rev,
                target_type=ObjectType.REVISION,
                message=None,
                metadata=None,
                synthetic=False,
                author=Person(name=None, email=None, fullname=b""),
                date=None,
            )
            missing_releases.add(release.id)
            releases[release.id] = release
            self.releases[name] = release.id

        if missing_releases:
            missing_releases = set(self.storage.release_missing(list(missing_releases)))

        for _id in missing_releases:
            yield releases[_id]

    def get_snapshot(self) -> Snapshot:
        """Get the snapshot that need to be loaded."""
        branches: Dict[bytes, Optional[SnapshotBranch]] = {}
        for name, (pointer_nature, target) in self.heads.items():
            branches[name] = SnapshotBranch(
                target=target, target_type=TargetType.REVISION
            )
            if pointer_nature == HEAD_POINTER_NAME:
                branches[b"HEAD"] = SnapshotBranch(
                    target=name, target_type=TargetType.ALIAS
                )
        for name, target in self.releases.items():
            branches[name] = SnapshotBranch(
                target=target, target_type=TargetType.RELEASE
            )

        self.snapshot = Snapshot(branches=branches)
        return self.snapshot

    def store_data(self) -> None:
        super().store_data()
        self.storage.extid_add(self.extids)

    def get_fetch_history_result(self):
        """Return the data to store in fetch_history."""
        return {
            "contents": self.num_contents,
            "directories": self.num_directories,
            "revisions": self.num_revisions,
            "releases": self.num_releases,
        }

    def load_status(self):
        snapshot = self.get_snapshot()
        load_status = "eventful"
        if self.last_snapshot_id is not None and self.last_snapshot_id == snapshot.id:
            load_status = "uneventful"
        return {
            "status": load_status,
        }


class HgArchiveBundle20Loader(HgBundle20Loader):
    """Mercurial loader for repository wrapped within archives.

    """

    def __init__(
        self,
        storage: StorageInterface,
        url: str,
        visit_date: Optional[datetime.datetime] = None,
        archive_path=None,
        temp_directory: str = "/tmp",
        max_content_size: Optional[int] = None,
    ):
        super().__init__(
            storage=storage,
            url=url,
            visit_date=visit_date,
            logging_class="swh.loader.mercurial.HgArchiveBundle20Loader",
            temp_directory=temp_directory,
            max_content_size=max_content_size,
        )
        self.archive_extract_temp_dir = None
        self.archive_path = archive_path

    def prepare(self):
        self.archive_extract_temp_dir = tmp_extract(
            archive=self.archive_path,
            dir=self.temp_directory,
            prefix=TEMPORARY_DIR_PREFIX_PATTERN,
            suffix=".dump-%s" % os.getpid(),
            log=self.log,
            source=self.origin_url,
        )

        repo_name = os.listdir(self.archive_extract_temp_dir)[0]
        self.directory = os.path.join(self.archive_extract_temp_dir, repo_name)
        super().prepare()



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
import signal
import time
import traceback

from mercurial import hg  # type: ignore
import pytest

from .. import hgutil


def test_clone_timeout(monkeypatch):
    src = "https://www.mercurial-scm.org/repo/hello"
    dest = "/dev/null"
    timeout = 0.1

    def clone(*args, **kwargs):
        # ignore SIGTERM to force sigkill
        signal.signal(signal.SIGTERM, lambda signum, frame: None)
        time.sleep(2)

    monkeypatch.setattr(hg, "clone", clone)

    with pytest.raises(hgutil.CloneTimeout) as e:
        hgutil.clone(src, dest, timeout)
    killed = True
    assert e.value.args == (src, timeout, killed)


def test_clone_error(caplog, tmp_path, monkeypatch):
    src = "https://www.mercurial-scm.org/repo/hello"
    dest = "/dev/null"
    expected_traceback = "Some traceback"

    def clone(*args, **kwargs):
        raise ValueError()

    def print_exc(file):
        file.write(expected_traceback)

    monkeypatch.setattr(hg, "clone", clone)
    monkeypatch.setattr(traceback, "print_exc", print_exc)

    with pytest.raises(hgutil.CloneFailure) as e:
        hgutil.clone(src, dest, 1)
    assert e.value.args == (src, dest, expected_traceback)



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime

import pytest

from swh.loader.mercurial.utils import parse_visit_date

VISIT_DATE_STR = "2021-02-17 15:50:04.518963"
VISIT_DATE = datetime(2021, 2, 17, 15, 50, 4, 518963)


@pytest.mark.parametrize(
    "input_visit_date,expected_date",
    [(None, None), (VISIT_DATE, VISIT_DATE), (VISIT_DATE_STR, VISIT_DATE),],
)
def test_utils_parse_visit_date(input_visit_date, expected_date):
    assert parse_visit_date(input_visit_date) == expected_date


def test_utils_parse_visit_date_now():
    actual_date = parse_visit_date("now")
    assert isinstance(actual_date, datetime)


def test_utils_parse_visit_date_fails():
    with pytest.raises(ValueError, match="invalid"):
        parse_visit_date(10)  # not a string nor a date



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import shutil
import tempfile

import patoolib


def tmp_extract(archive, dir=None, prefix=None, suffix=None, log=None, source=None):
    """Extract an archive to a temporary location with optional logs.

    Args:
        archive (string): Absolute path of the archive to be extracted
        prefix (string): Optional modifier to the temporary storage
            directory name. (I guess in case something
            goes wrong and you want to go look?)
        log (python logging instance): Optional for recording extractions.
        source (string): Optional source URL of the archive for adding to
            log messages.
    Returns:
        A context manager for a temporary directory that automatically
        removes itself. See: help(tempfile.TemporaryDirectory)
    """
    logstr = "From %s - " % source if log and source else ""

    if dir and not os.path.exists(dir):
        os.makedirs(dir, exist_ok=True)

    archive_base = os.path.basename(archive)
    if archive_base[0] == ".":
        package = "." + archive_base.split(".")[1]
    else:
        package = archive_base.split(".")[0]

    tmpdir = tempfile.mkdtemp(dir=dir, prefix=prefix, suffix=suffix)
    repo_path = os.path.join(tmpdir, package)
    try:
        patoolib.extract_archive(archive, interactive=False, outdir=tmpdir)
    except Exception as e:
        if os.path.exists(tmpdir):
            shutil.rmtree(tmpdir)
        msg = "%sFailed to uncompress archive %s at %s - %s" % (
            logstr,
            archive_base,
            repo_path,
            e,
        )
        raise ValueError(msg)

    if log:
        log.info("%sUncompressing archive %s at %s" % (logstr, archive_base, repo_path))
    return tmpdir



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
from datetime import datetime
from hashlib import sha1
import os
from pathlib import Path
import subprocess

import attr
import pytest

from swh.loader.mercurial.loader import HgBundle20Loader
from swh.loader.mercurial.utils import parse_visit_date
from swh.loader.tests import (
    assert_last_visit_matches,
    check_snapshot,
    get_stats,
    prepare_repository_from_archive,
)
from swh.model.from_disk import Content, DentryPerms
from swh.model.hashutil import hash_to_bytes, hash_to_hex
from swh.model.identifiers import ObjectType
from swh.model.model import RevisionType, Snapshot, SnapshotBranch, TargetType
from swh.storage import get_storage
from swh.storage.algos.snapshot import snapshot_get_latest

from ..from_disk import HgDirectory, HgLoaderFromDisk
from .loader_checker import ExpectedSwhids, LoaderChecker

VISIT_DATE = parse_visit_date("2016-05-03 15:16:32+00")
assert VISIT_DATE is not None


def random_content() -> Content:
    """Create minimal content object."""
    data = str(datetime.now()).encode()
    return Content({"sha1_git": sha1(data).digest(), "perms": DentryPerms.content})


def test_hg_directory_creates_missing_directories():
    directory = HgDirectory()
    directory[b"path/to/some/content"] = random_content()


def test_hg_directory_get():
    content = random_content()
    directory = HgDirectory()

    assert directory.get(b"path/to/content") is None
    assert directory.get(b"path/to/content", content) == content

    directory[b"path/to/content"] = content
    assert directory.get(b"path/to/content") == content


def test_hg_directory_deletes_empty_directories():
    directory = HgDirectory()
    content = random_content()
    directory[b"path/to/content"] = content
    directory[b"path/to/some/deep/content"] = random_content()

    del directory[b"path/to/some/deep/content"]

    assert directory.get(b"path/to/some/deep") is None
    assert directory.get(b"path/to/some") is None
    assert directory.get(b"path/to/content") == content


def test_hg_directory_when_directory_replaces_file():
    directory = HgDirectory()
    directory[b"path/to/some"] = random_content()
    directory[b"path/to/some/content"] = random_content()


# Those tests assert expectations on repository loading
# by reading expected values from associated json files
# produced by the `swh-hg-identify` command line utility.
#
# It has more granularity than historical tests.
# Assertions will tell if the error comes from the directories
# revisions or release rather than only checking the snapshot.
#
# With more work it should event be possible to know which part
# of an object is faulty.
@pytest.mark.parametrize(
    "archive_name", ("hello", "transplant", "the-sandbox", "example")
)
def test_examples(swh_storage, datadir, tmp_path, archive_name):
    archive_path = Path(datadir, f"{archive_name}.tgz")
    json_path = Path(datadir, f"{archive_name}.json")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    LoaderChecker(
        loader=HgLoaderFromDisk(swh_storage, repo_url),
        expected=ExpectedSwhids.load(json_path),
    ).check()


# This test has as been adapted from the historical `HgBundle20Loader` tests
# to ensure compatibility of `HgLoaderFromDisk`.
# Hashes as been produced by copy pasting the result of the implementation
# to prevent regressions.
def test_loader_hg_new_visit_no_release(swh_storage, datadir, tmp_path):
    """Eventful visit should yield 1 snapshot"""
    archive_name = "the-sandbox"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = HgLoaderFromDisk(swh_storage, url=repo_url)

    assert loader.load() == {"status": "eventful"}

    tips = {
        b"branch-tip/default": "70e750bb046101fdced06f428e73fee471509c56",
        b"branch-tip/develop": "a9c4534552df370f43f0ef97146f393ef2f2a08c",
    }
    closed = {
        b"feature/fun_time": "4d640e8064fe69b4c851dfd43915c431e80c7497",
        b"feature/green2_loader": "94be9abcf9558213ff301af0ecd8223451ce991d",
        b"feature/greenloader": "9f82d95bd3edfb7f18b1a21d6171170395ea44ce",
        b"feature/my_test": "dafa445964230e808148db043c126063ea1dc9b6",
        b"feature/read2_loader": "9e912851eb64e3a1e08fbb587de7a4c897ce5a0a",
        b"feature/readloader": "ddecbc16f4c916c39eacfcb2302e15a9e70a231e",
        b"feature/red": "cb36b894129ca7910bb81c457c72d69d5ff111bc",
        b"feature/split5_loader": "3ed4b85d30401fe32ae3b1d650f215a588293a9e",
        b"feature/split_causing": "c346f6ff7f42f2a8ff867f92ab83a6721057d86c",
        b"feature/split_loader": "5f4eba626c3f826820c4475d2d81410759ec911b",
        b"feature/split_loader5": "5017ce0b285351da09a2029ea2cf544f79b593c7",
        b"feature/split_loading": "4e2dc6d6073f0b6d348f84ded52f9143b10344b9",
        b"feature/split_redload": "2d4a801c9a9645fcd3a9f4c06418d8393206b1f3",
        b"feature/splitloading": "88b80615ed8561be74a700b92883ec0374ddacb0",
        b"feature/test": "61d762d65afb3150e2653d6735068241779c1fcf",
        b"feature/test_branch": "be44d5e6cc66580f59c108f8bff5911ee91a22e4",
        b"feature/test_branching": "d2164061453ecb03d4347a05a77db83f706b8e15",
        b"feature/test_dog": "2973e5dc9568ac491b198f6b7f10c44ddc04e0a3",
    }

    mapping = {b"branch-closed-heads/%s/0" % b: n for b, n in closed.items()}
    mapping.update(tips)

    expected_branches = {
        k: SnapshotBranch(target=hash_to_bytes(v), target_type=TargetType.REVISION)
        for k, v in mapping.items()
    }
    expected_branches[b"HEAD"] = SnapshotBranch(
        target=b"branch-tip/default", target_type=TargetType.ALIAS
    )

    expected_snapshot = Snapshot(
        id=hash_to_bytes("cbc609dcdced34dbd9938fe81b555170f1abc96f"),
        branches=expected_branches,
    )

    assert_last_visit_matches(
        loader.storage,
        repo_url,
        status="full",
        type="hg",
        snapshot=expected_snapshot.id,
    )
    check_snapshot(expected_snapshot, loader.storage)

    stats = get_stats(loader.storage)
    assert stats == {
        "content": 2,
        "directory": 3,
        "origin": 1,
        "origin_visit": 1,
        "release": 0,
        "revision": 58,
        "skipped_content": 0,
        "snapshot": 1,
    }


# This test has as been adapted from the historical `HgBundle20Loader` tests
# to ensure compatibility of `HgLoaderFromDisk`.
# Hashes as been produced by copy pasting the result of the implementation
# to prevent regressions.
def test_loader_hg_new_visit_with_release(swh_storage, datadir, tmp_path):
    """Eventful visit with release should yield 1 snapshot"""

    archive_name = "hello"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = HgLoaderFromDisk(swh_storage, url=repo_url, visit_date=VISIT_DATE,)

    actual_load_status = loader.load()
    assert actual_load_status == {"status": "eventful"}

    # then
    stats = get_stats(loader.storage)
    assert stats == {
        "content": 3,
        "directory": 3,
        "origin": 1,
        "origin_visit": 1,
        "release": 1,
        "revision": 3,
        "skipped_content": 0,
        "snapshot": 1,
    }

    # cf. test_loader.org for explaining from where those hashes
    tip_release = hash_to_bytes("515c4d72e089404356d0f4b39d60f948b8999140")
    release = loader.storage.release_get([tip_release])[0]
    assert release is not None

    tip_revision_default = hash_to_bytes("c3dbe4fbeaaa98dd961834e4007edb3efb0e2a27")
    revision = loader.storage.revision_get([tip_revision_default])[0]
    assert revision is not None

    expected_snapshot = Snapshot(
        id=hash_to_bytes("7ef082aa8b53136b1bed97f734504be32679bbec"),
        branches={
            b"branch-tip/default": SnapshotBranch(
                target=tip_revision_default, target_type=TargetType.REVISION,
            ),
            b"tags/0.1": SnapshotBranch(
                target=tip_release, target_type=TargetType.RELEASE,
            ),
            b"HEAD": SnapshotBranch(
                target=b"branch-tip/default", target_type=TargetType.ALIAS,
            ),
        },
    )

    check_snapshot(expected_snapshot, loader.storage)
    assert_last_visit_matches(
        loader.storage,
        repo_url,
        type=RevisionType.MERCURIAL.value,
        status="full",
        snapshot=expected_snapshot.id,
    )


# This test has as been adapted from the historical `HgBundle20Loader` tests
# to ensure compatibility of `HgLoaderFromDisk`.
# Hashes as been produced by copy pasting the result of the implementation
# to prevent regressions.
def test_visit_repository_with_transplant_operations(swh_storage, datadir, tmp_path):
    """Visit a mercurial repository visit transplant operations within should yield a
    snapshot as well.

    """

    archive_name = "transplant"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = HgLoaderFromDisk(swh_storage, url=repo_url, visit_date=VISIT_DATE,)

    # load hg repository
    actual_load_status = loader.load()
    assert actual_load_status == {"status": "eventful"}

    # collect swh revisions
    assert_last_visit_matches(
        loader.storage, repo_url, type=RevisionType.MERCURIAL.value, status="full"
    )

    revisions = []
    snapshot = snapshot_get_latest(loader.storage, repo_url)
    for branch in snapshot.branches.values():
        if branch.target_type.value != "revision":
            continue
        revisions.append(branch.target)

    # extract original changesets info and the transplant sources
    hg_changesets = set()
    transplant_sources = set()
    for rev in loader.storage.revision_log(revisions):
        extids = list(
            loader.storage.extid_get_from_target(ObjectType.REVISION, [rev["id"]])
        )
        assert len(extids) == 1
        hg_changesets.add(hash_to_hex(extids[0].extid))
        for k, v in rev["extra_headers"]:
            if k == b"transplant_source":
                transplant_sources.add(v.decode("ascii"))

    # check extracted data are valid
    assert len(hg_changesets) > 0
    assert len(transplant_sources) > 0
    assert transplant_sources <= hg_changesets


def _partial_copy_storage(
    old_storage, origin_url: str, mechanism: str, copy_revisions: bool
):
    """Create a new storage, and only copy ExtIDs or head revisions to it."""
    new_storage = get_storage(cls="memory")
    snapshot = snapshot_get_latest(old_storage, origin_url)
    assert snapshot
    heads = [branch.target for branch in snapshot.branches.values()]

    if mechanism == "extid":
        extids = old_storage.extid_get_from_target(ObjectType.REVISION, heads)
        new_storage.extid_add(extids)
        if copy_revisions:
            # copy revisions, but erase their metadata to make sure the loader doesn't
            # fallback to revision.metadata["nodeid"]
            revisions = [
                attr.evolve(rev, metadata={})
                for rev in old_storage.revision_get(heads)
                if rev
            ]
            new_storage.revision_add(revisions)

    else:
        assert mechanism == "same storage"
        return old_storage

    # copy origin, visit, status
    new_storage.origin_add(old_storage.origin_get([origin_url]))
    visit = old_storage.origin_visit_get_latest(origin_url)
    new_storage.origin_visit_add([visit])
    statuses = old_storage.origin_visit_status_get(origin_url, visit.visit).results
    new_storage.origin_visit_status_add(statuses)
    new_storage.snapshot_add([snapshot])

    return new_storage


@pytest.mark.parametrize("mechanism", ("extid", "same storage"))
def test_load_unchanged_repo_should_be_uneventful(
    swh_storage, datadir, tmp_path, mechanism
):
    """Checks the loader can find which revisions it already loaded, using ExtIDs."""
    archive_name = "hello"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)
    repo_path = repo_url.replace("file://", "")

    loader = HgLoaderFromDisk(swh_storage, repo_path)

    assert loader.load() == {"status": "eventful"}
    assert get_stats(loader.storage) == {
        "content": 3,
        "directory": 3,
        "origin": 1,
        "origin_visit": 1,
        "release": 1,
        "revision": 3,
        "skipped_content": 0,
        "snapshot": 1,
    }

    old_storage = swh_storage

    # Create a new storage, and only copy ExtIDs or head revisions to it.
    # This should be enough for the loader to know revisions were already loaded
    new_storage = _partial_copy_storage(
        old_storage, repo_path, mechanism=mechanism, copy_revisions=True
    )

    # Create a new loader (to start with a clean slate, eg. remove the caches),
    # with the new, partial, storage
    loader = HgLoaderFromDisk(new_storage, repo_path)
    assert loader.load() == {"status": "uneventful"}

    if mechanism == "same storage":
        # Should have all the objects
        assert get_stats(loader.storage) == {
            "content": 3,
            "directory": 3,
            "origin": 1,
            "origin_visit": 2,
            "release": 1,
            "revision": 3,
            "skipped_content": 0,
            "snapshot": 1,
        }
    else:
        # Should have only the objects we directly inserted from the test, plus
        # a new visit
        assert get_stats(loader.storage) == {
            "content": 0,
            "directory": 0,
            "origin": 1,
            "origin_visit": 2,
            "release": 0,
            "revision": 1,
            "skipped_content": 0,
            "snapshot": 1,
        }


def test_closed_branch_incremental(swh_storage, datadir, tmp_path):
    """Test that a repository with a closed branch does not trip an incremental load"""
    archive_name = "example"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)
    repo_path = repo_url.replace("file://", "")

    loader = HgLoaderFromDisk(swh_storage, repo_path)

    # Test 3 loads: full, and two incremental.
    assert loader.load() == {"status": "eventful"}
    expected_stats = {
        "content": 7,
        "directory": 16,
        "origin": 1,
        "origin_visit": 1,
        "release": 0,
        "revision": 9,
        "skipped_content": 0,
        "snapshot": 1,
    }
    assert get_stats(loader.storage) == expected_stats
    assert loader.load() == {"status": "uneventful"}
    assert get_stats(loader.storage) == {**expected_stats, "origin_visit": 1 + 1}
    assert loader.load() == {"status": "uneventful"}
    assert get_stats(loader.storage) == {**expected_stats, "origin_visit": 2 + 1}


def test_old_loader_new_loader(swh_storage, datadir, tmp_path):
    archive_name = "example"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)
    repo_path = repo_url.replace("file://", "")

    old_loader = HgBundle20Loader(swh_storage, repo_path)
    assert old_loader.load() == {"status": "eventful"}

    expected_stats = {
        "content": 7,
        "directory": 16,
        "origin": 1,
        "origin_visit": 1,
        "release": 0,
        "revision": 9,
        "skipped_content": 0,
        "snapshot": 1,
    }
    assert get_stats(old_loader.storage) == expected_stats

    # Will pick up more branches, hence a different snapshot
    loader = HgLoaderFromDisk(swh_storage, repo_path)
    res = loader.load()
    new_expected_stats = {
        **expected_stats,
        "origin_visit": 2,
        "snapshot": 2,
    }
    assert get_stats(loader.storage) == new_expected_stats
    assert res == {"status": "eventful"}

    # Shouldn't pick up anything now
    loader = HgLoaderFromDisk(swh_storage, repo_path)
    assert loader.load() == {"status": "uneventful"}

    # Shouldn't pick up anything either after another load
    loader = HgLoaderFromDisk(swh_storage, repo_path)
    assert loader.load() == {"status": "uneventful"}


def test_load_unchanged_repo__dangling_extid(swh_storage, datadir, tmp_path):
    """Checks the loader will load revisions targeted by an ExtID if the
    revisions are missing from the storage"""
    archive_name = "hello"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)
    repo_path = repo_url.replace("file://", "")

    loader = HgLoaderFromDisk(swh_storage, repo_path)

    assert loader.load() == {"status": "eventful"}
    assert get_stats(loader.storage) == {
        "content": 3,
        "directory": 3,
        "origin": 1,
        "origin_visit": 1,
        "release": 1,
        "revision": 3,
        "skipped_content": 0,
        "snapshot": 1,
    }

    old_storage = swh_storage

    # Create a new storage, and only copy ExtIDs or head revisions to it.
    # This should be enough for the loader to know revisions were already loaded
    new_storage = _partial_copy_storage(
        old_storage, repo_path, mechanism="extid", copy_revisions=False
    )

    # Create a new loader (to start with a clean slate, eg. remove the caches),
    # with the new, partial, storage
    loader = HgLoaderFromDisk(new_storage, repo_path)

    assert get_stats(loader.storage) == {
        "content": 0,
        "directory": 0,
        "origin": 1,
        "origin_visit": 1,
        "release": 0,
        "revision": 0,
        "skipped_content": 0,
        "snapshot": 1,
    }

    assert loader.load() == {"status": "eventful"}

    assert get_stats(loader.storage) == {
        "content": 3,
        "directory": 3,
        "origin": 1,
        "origin_visit": 2,
        "release": 1,
        "revision": 3,
        "skipped_content": 0,
        "snapshot": 1,
    }


def test_missing_filelog_should_not_crash(swh_storage, datadir, tmp_path):
    archive_name = "missing-filelog"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)
    directory = repo_url.replace("file://", "")

    loader = HgLoaderFromDisk(
        storage=swh_storage,
        url=repo_url,
        directory=directory,  # specify directory to avoid clone
        visit_date=VISIT_DATE,
    )

    actual_load_status = loader.load()
    assert actual_load_status == {"status": "eventful"}

    assert_last_visit_matches(swh_storage, repo_url, status="partial", type="hg")


def test_multiple_open_heads(swh_storage, datadir, tmp_path):
    archive_name = "multiple-heads"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    loader = HgLoaderFromDisk(storage=swh_storage, url=repo_url,)

    actual_load_status = loader.load()
    assert actual_load_status == {"status": "eventful"}

    assert_last_visit_matches(swh_storage, repo_url, status="full", type="hg")

    snapshot = snapshot_get_latest(swh_storage, repo_url)
    expected_branches = [
        b"HEAD",
        b"branch-heads/default/0",
        b"branch-heads/default/1",
        b"branch-tip/default",
    ]
    assert sorted(snapshot.branches.keys()) == expected_branches

    # Check that we don't load anything the second time
    loader = HgLoaderFromDisk(storage=swh_storage, url=repo_url,)

    actual_load_status = loader.load()

    assert actual_load_status == {"status": "uneventful"}


def hg_strip(repo: str, revset: str) -> None:
    """Removes `revset` and all of their descendants from the local repository."""
    # Previously called `hg strip`, it was renamed to `hg debugstrip` in Mercurial 5.7
    # because it's most likely not what most users want to do (they should use some kind
    # of history-rewriting tool like `histedit` or `prune`).
    # But here, it's exactly what we want to do.
    subprocess.check_call(["hg", "debugstrip", revset], cwd=repo)


def test_load_repo_with_new_commits(swh_storage, datadir, tmp_path):
    archive_name = "hello"
    archive_path = Path(datadir, f"{archive_name}.tgz")
    json_path = Path(datadir, f"{archive_name}.json")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)

    # first load with missing commits
    hg_strip(repo_url.replace("file://", ""), "tip")
    loader = HgLoaderFromDisk(swh_storage, repo_url)
    assert loader.load() == {"status": "eventful"}
    assert get_stats(loader.storage) == {
        "content": 2,
        "directory": 2,
        "origin": 1,
        "origin_visit": 1,
        "release": 0,
        "revision": 2,
        "skipped_content": 0,
        "snapshot": 1,
    }

    # second load with all commits
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)
    loader = HgLoaderFromDisk(swh_storage, repo_url)
    checker = LoaderChecker(loader=loader, expected=ExpectedSwhids.load(json_path),)

    checker.check()

    assert get_stats(loader.storage) == {
        "content": 3,
        "directory": 3,
        "origin": 1,
        "origin_visit": 2,
        "release": 1,
        "revision": 3,
        "skipped_content": 0,
        "snapshot": 2,
    }



# Copyright (C) 2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import struct


class ChunkedFileReader(object):
    """A binary stream reader that gives seamless read access to Mercurial's
    bundle2 HG20 format which is partitioned for some reason at the file level
    into chunks of [4Bytes:<length>, <length>Bytes:<data>] as if it were
    encoding transport packets.

    args:
        file: rb file handle pre-aligned to the start of the chunked portion
        size_unpack_fmt: struct format string for unpacking the next chunk size
    """

    def __init__(self, file, size_unpack_fmt=">I"):
        self._file = file
        self._size_pattern = size_unpack_fmt
        self._size_bytes = struct.calcsize(size_unpack_fmt)
        self._bytes_per_chunk = self._chunk_size(True)
        self._chunk_bytes_left = self._bytes_per_chunk
        self._offset = self._file.tell()
        # find the file size
        self._file.seek(0, 2)  # seek to end
        self._size = self._file.tell()
        self._file.seek(self._offset, 0)  # seek back to original position

    def _chunk_size(self, first_time=False):
        """Unpack the next bytes from the file to get the next file chunk size.
        """
        size = struct.unpack(self._size_pattern, self._file.read(self._size_bytes))[0]
        return size

    def size(self):
        """Returns the file size in bytes.
        """
        return self._size

    def read(self, bytes_to_read):
        """Return N bytes from the file as a single block.

        args:
            bytes_to_read: int number of bytes of content
        """
        return b"".join(self.read_iterator(bytes_to_read))

    def read_iterator(self, bytes_to_read):
        """Return a generator that yields N bytes from the file one file chunk
        at a time.

        args:
            bytes_to_read: int number of bytes of content
        """
        while bytes_to_read > self._chunk_bytes_left:
            yield self._file.read(self._chunk_bytes_left)
            bytes_to_read -= self._chunk_bytes_left
            self._chunk_bytes_left = self._chunk_size()
        self._chunk_bytes_left -= bytes_to_read
        yield self._file.read(bytes_to_read)

    def seek(self, new_pos=None, from_current=False):
        """Wraps the underlying file seek, additionally updating the
        chunk_bytes_left counter appropriately so that we can start reading
        from the new location.

        args:

            new_pos: new cursor byte position
            from_current: if True, it treats new_pos as an offset from the
                current cursor position, bypassing any chunk boundaries as if
                they weren't there. This should give the same end position as a
                read except without the reading data part.
        """
        if from_current:
            new_pos = new_pos or 0  # None -> current
            if new_pos <= self._chunk_bytes_left:
                new_pos += self._file.tell()
            else:
                new_pos += (
                    self._file.tell()
                    + self._size_bytes
                    + (
                        (
                            (new_pos - self._chunk_bytes_left - 1)  # aliasing
                            // self._bytes_per_chunk
                        )
                        * self._size_bytes
                    )
                )
        else:
            new_pos = new_pos or self._offset  # None -> start position

        self._chunk_bytes_left = self._bytes_per_chunk - (
            (new_pos - self._offset) % (self._bytes_per_chunk + self._size_bytes)
        )
        self._file.seek(new_pos)

    def __getattr__(self, item):
        """Forward other calls to the underlying file object.
        """
        return getattr(self._file, item)



# Copyright (C) 2015-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


PRIMARY_ALGO = "sha1_git"


def parse_author(name_email):
    """Parse an author line"""

    if name_email is None:
        return None

    try:
        open_bracket = name_email.index(b"<")
    except ValueError:
        name = email = None
    else:
        raw_name = name_email[:open_bracket]
        raw_email = name_email[open_bracket + 1 :]

        if not raw_name:
            name = None
        elif raw_name.endswith(b" "):
            name = raw_name[:-1]
        else:
            name = raw_name

        try:
            close_bracket = raw_email.index(b">")
        except ValueError:
            email = None
        else:
            email = raw_email[:close_bracket]

    return {
        "name": name,
        "email": email,
        "fullname": name_email,
    }



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from textwrap import dedent
from urllib.parse import urlsplit

from click.testing import CliRunner

from swh.loader.mercurial.identify import main
from swh.loader.tests import prepare_repository_from_archive


def test_all_revisions(datadir: str, tmp_path: str):
    archive_name = "hello"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)
    directory = urlsplit(repo_url).path

    runner = CliRunner()
    result = runner.invoke(main, ["-d", directory, "revision"])

    expected = dedent(
        """
        swh:1:rev:93b48d515580522a05f389bec93227fc8e43d940\t0a04b987be5ae354b710cefeba0e2d9de7ad41a9
        swh:1:rev:8dd3db5d5519e4947f035d141581d304565372d2\t82e55d328c8ca4ee16520036c0aaace03a5beb65
        swh:1:rev:c3dbe4fbeaaa98dd961834e4007edb3efb0e2a27\tb985ae4a07e12ac662f45a171e2d42b13be5b50c
        """
    ).lstrip()
    assert result.output == expected


def test_single_revision(datadir: str, tmp_path: str):
    archive_name = "hello"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)
    directory = urlsplit(repo_url).path

    runner = CliRunner()
    result = runner.invoke(
        main, ["-d", directory, "revision", "0a04b987be5ae354b710cefeba0e2d9de7ad41a9"]
    )

    expected = (
        "swh:1:rev:93b48d515580522a05f389bec93227fc8e43d940"
        "\t0a04b987be5ae354b710cefeba0e2d9de7ad41a9\n"
    )
    assert result.output == expected


def test_all(datadir: str, tmp_path: str):
    archive_name = "hello"
    archive_path = os.path.join(datadir, f"{archive_name}.tgz")
    repo_url = prepare_repository_from_archive(archive_path, archive_name, tmp_path)
    directory = urlsplit(repo_url).path

    runner = CliRunner()
    result = runner.invoke(main, ["-d", directory, "all"])

    expected = dedent(
        f"""
        swh:1:dir:43d727f2f3f2f7cb3b098ddad1d7038464a4cee2\t0a04b987be5ae354b710cefeba0e2d9de7ad41a9
        swh:1:dir:b3f85f210ff86d334575f64cb01c5bf49895b63e\t82e55d328c8ca4ee16520036c0aaace03a5beb65
        swh:1:dir:8f2be433c945384c85920a8e60f2a68d2c0f20fb\tb985ae4a07e12ac662f45a171e2d42b13be5b50c
        swh:1:rev:93b48d515580522a05f389bec93227fc8e43d940\t0a04b987be5ae354b710cefeba0e2d9de7ad41a9
        swh:1:rev:8dd3db5d5519e4947f035d141581d304565372d2\t82e55d328c8ca4ee16520036c0aaace03a5beb65
        swh:1:rev:c3dbe4fbeaaa98dd961834e4007edb3efb0e2a27\tb985ae4a07e12ac662f45a171e2d42b13be5b50c
        swh:1:rel:515c4d72e089404356d0f4b39d60f948b8999140\t0.1
        swh:1:snp:d35668e02e2ba4321dc951cd308cf883786f918a\t{directory}
        """
    ).lstrip()
    assert result.output == expected



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
from collections import defaultdict
from dataclasses import dataclass
import io
import os
import signal
import time
import traceback
from typing import Dict, List, Mapping, NewType, Optional, Set

from billiard import Process, Queue

# The internal Mercurial API is not guaranteed to be stable.
from mercurial import bookmarks, context, error, hg, smartset, util  # type: ignore
import mercurial.ui  # type: ignore

NULLID = mercurial.node.nullid
HgNodeId = NewType("HgNodeId", bytes)
Repository = hg.localrepo
BaseContext = context.basectx
LRUCacheDict = util.lrucachedict
HgSpanSet = smartset._spanset
HgFilteredSet = smartset.filteredset
LookupError = error.LookupError


def repository(path: str) -> hg.localrepo:
    ui = mercurial.ui.ui.load()
    return hg.repository(ui, path.encode())


@dataclass
class BranchingInfo:
    tips: Mapping[bytes, HgNodeId]
    """The first head of the branch, sorted by nodeid if there are multiple heads."""
    bookmarks: Mapping[bytes, HgNodeId]
    """all bookmarks in the repository (except local divergent ones)"""
    open_heads: Mapping[bytes, List[HgNodeId]]
    """All *open* heads of a given branch, sorted by nodeid"""
    closed_heads: Mapping[bytes, List[HgNodeId]]
    """All *closed* heads of a given branch, sorted by nodeid, if any"""
    default_branch_alias: Optional[bytes]
    """The default snapshot branch to show in the UI"""


def branching_info(repo: hg.localrepo, ignored: Set[int]) -> BranchingInfo:
    """Lists all relevant information about branch heads and bookmarks, grouped by type.

    `ignored`: Revisions that we ignore during loading because they are corrupted or
    have a corrupted ancestor.

    Categories may have overlapping nodes: a branch tip can be a closed branch head
    and have a bookmark on it, for example.
    """
    branch_tips: Dict[bytes, HgNodeId] = {}
    branch_open_heads = defaultdict(list)
    branch_closed_heads = defaultdict(list)
    all_bookmarks = bookmarks.listbookmarks(repo)

    for branch_name, heads in repo.branchmap().items():
        # Sort the heads by node id since it's stable and doesn't depend on local
        # topology like cloning order.
        for head in sorted(heads):
            head = repo[head]
            if head.rev() in ignored:
                # This revision or one of its ancestors is corrupted, ignore it
                continue
            node_id = head.node()
            if head.closesbranch():
                branch_closed_heads[branch_name].append(node_id)
            else:
                if not branch_tips.get(branch_name):
                    branch_tips[branch_name] = node_id
                branch_open_heads[branch_name].append(node_id)

    # The default revision is where the "@" bookmark is, or failing that the tip of the
    # `default` branch. For our purposes we're trying to find a branch tip to alias to,
    # so only return those if they are branch tips, otherwise don't bother.
    default_rev_alias = None
    at_bookmark = all_bookmarks.get(b"@")
    if at_bookmark is not None:
        bookmark_at_branch = repo[at_bookmark].branch()
        if branch_tips.get(bookmark_at_branch) is not None:
            default_rev_alias = b"bookmarks/@"
    if default_rev_alias is None and branch_tips.get(b"default") is not None:
        default_rev_alias = b"branch-tip/default"

    branches_with_one_head = set()
    for branch, heads in branch_open_heads.items():
        if len(heads) == 1:
            branches_with_one_head.add(branch)

    # The most common case is one head per branch. Simplifying this means we have
    # less duplicate data, because open heads are the same as open branch tips.
    # We don't do more complex deduplication, this is just a simple optimization.
    for branch in branches_with_one_head:
        del branch_open_heads[branch]

    return BranchingInfo(
        tips=branch_tips,
        bookmarks=all_bookmarks,
        open_heads=branch_open_heads,
        closed_heads=branch_closed_heads,
        default_branch_alias=default_rev_alias,
    )


class CloneTimeout(Exception):
    pass


class CloneFailure(Exception):
    pass


def _clone_task(src: str, dest: str, errors: Queue) -> None:
    """Clone task to run in a subprocess.

    Args:
        src: clone source
        dest: clone destination
        errors: message queue to communicate errors
    """
    try:
        hg.clone(
            ui=mercurial.ui.ui.load(),
            peeropts={},
            source=src.encode(),
            dest=dest.encode(),
            update=False,
        )
    except Exception as e:
        exc_buffer = io.StringIO()
        traceback.print_exc(file=exc_buffer)
        errors.put_nowait(exc_buffer.getvalue())
        raise e


def clone(src: str, dest: str, timeout: float) -> None:
    """Clone a repository with timeout.

    Args:
        src: clone source
        dest: clone destination
        timeout: timeout in seconds
    """
    errors: Queue = Queue()
    process = Process(target=_clone_task, args=(src, dest, errors))
    process.start()
    process.join(timeout)

    if process.is_alive():
        process.terminate()
        # Give it a second (literally), then kill it
        # Can't use `process.join(1)` here, billiard appears to be bugged
        # https://github.com/celery/billiard/issues/270
        killed = False
        for _ in range(10):
            time.sleep(0.1)
            if not process.is_alive():
                break
        else:
            killed = True
            os.kill(process.pid, signal.SIGKILL)
        raise CloneTimeout(src, timeout, killed)

    if not errors.empty():
        raise CloneFailure(src, dest, errors.get())



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import deque
from datetime import datetime
import os
from shutil import rmtree
from tempfile import mkdtemp
from typing import Deque, Dict, List, Optional, Set, Tuple, TypeVar, Union

from swh.loader.core.loader import BaseLoader
from swh.loader.core.utils import clean_dangling_folders
from swh.loader.mercurial.utils import get_minimum_env, parse_visit_date
from swh.model import identifiers
from swh.model.from_disk import Content, DentryPerms, Directory
from swh.model.hashutil import hash_to_bytehex
from swh.model.model import (
    ExtID,
    ObjectType,
    Origin,
    Person,
    Release,
    Revision,
    RevisionType,
    Sha1Git,
    Snapshot,
    SnapshotBranch,
    TargetType,
    TimestampWithTimezone,
)
from swh.model.model import Content as ModelContent
from swh.storage.algos.snapshot import snapshot_get_latest
from swh.storage.interface import StorageInterface

from . import hgutil
from .archive_extract import tmp_extract
from .hgutil import HgFilteredSet, HgNodeId, HgSpanSet

FLAG_PERMS = {
    b"l": DentryPerms.symlink,
    b"x": DentryPerms.executable_content,
    b"": DentryPerms.content,
}  # type: Dict[bytes, DentryPerms]

TEMPORARY_DIR_PREFIX_PATTERN = "swh.loader.mercurial.from_disk"

EXTID_TYPE = "hg-nodeid"


T = TypeVar("T")


class CorruptedRevision(ValueError):
    """Raised when a revision is corrupted."""

    def __init__(self, hg_nodeid: HgNodeId) -> None:
        super().__init__(hg_nodeid.hex())
        self.hg_nodeid = hg_nodeid


class HgDirectory(Directory):
    """A more practical directory.

    - creates missing parent directories
    - removes empty directories
    """

    def __setitem__(self, path: bytes, value: Union[Content, "HgDirectory"]) -> None:
        if b"/" in path:
            head, tail = path.split(b"/", 1)

            directory = self.get(head)
            if directory is None or isinstance(directory, Content):
                directory = HgDirectory()
                self[head] = directory

            directory[tail] = value
        else:
            super().__setitem__(path, value)

    def __delitem__(self, path: bytes) -> None:
        super().__delitem__(path)

        while b"/" in path:  # remove empty parent directories
            path = path.rsplit(b"/", 1)[0]
            if len(self[path]) == 0:
                super().__delitem__(path)
            else:
                break

    def get(
        self, path: bytes, default: Optional[T] = None
    ) -> Optional[Union[Content, "HgDirectory", T]]:
        # TODO move to swh.model.from_disk.Directory
        try:
            return self[path]
        except KeyError:
            return default


class HgLoaderFromDisk(BaseLoader):
    """Load a mercurial repository from a local repository.

    Mercurial's branching model is more complete than Git's; it allows for multiple
    heads per branch, closed heads and bookmarks. The following mapping is used to
    represent the branching state of a Mercurial project in a given snapshot:

    - `HEAD` (optional) either the node pointed by the `@` bookmark or the tip of
      the `default` branch
    - `branch-tip/<branch-name>` (required) the first head of the branch, sorted by
      nodeid if there are multiple heads.
    - `bookmarks/<bookmark_name>` (optional) holds the bookmarks mapping if any
    - `branch-heads/<branch_name>/0..n` (optional) for any branch with multiple open
      heads, list all *open* heads
    - `branch-closed-heads/<branch_name>/0..n` (optional) for any branch with at least
      one closed head, list all *closed* heads
    - `tags/<tag-name>` (optional) record tags

    The format is not ambiguous regardless of branch name since we know it ends with a
    `/<index>`, as long as we have a stable sorting of the heads (we sort by nodeid).
    There may be some overlap between the refs, but it's simpler not to try to figure
    out de-duplication.
    However, to reduce the redundancy between snapshot branches in the most common case,
    when a branch has a single open head, it will only be referenced as
    `branch-tip/<branch-name>`. The `branch-heads/` hierarchy only appears when a branch
    has multiple open heads, which we consistently sort by increasing nodeid.
    The `branch-closed-heads/` hierarchy is also sorted by increasing nodeid.
    """

    CONFIG_BASE_FILENAME = "loader/mercurial"

    visit_type = "hg"

    def __init__(
        self,
        storage: StorageInterface,
        url: str,
        directory: Optional[str] = None,
        logging_class: str = "swh.loader.mercurial.LoaderFromDisk",
        visit_date: Optional[datetime] = None,
        temp_directory: str = "/tmp",
        clone_timeout_seconds: int = 7200,
        content_cache_size: int = 10_000,
        max_content_size: Optional[int] = None,
    ):
        """Initialize the loader.

        Args:
            url: url of the repository.
            directory: directory of the local repository.
            logging_class: class of the loader logger.
            visit_date: visit date of the repository
            config: loader configuration
        """
        super().__init__(
            storage=storage,
            logging_class=logging_class,
            max_content_size=max_content_size,
        )

        self._temp_directory = temp_directory
        self._clone_timeout = clone_timeout_seconds

        self.origin_url = url
        self.visit_date = visit_date
        self.directory = directory

        self._repo: Optional[hgutil.Repository] = None
        self._revision_nodeid_to_sha1git: Dict[HgNodeId, Sha1Git] = {}
        self._repo_directory: Optional[str] = None

        # keeps the last processed hg nodeid
        # it is used for differential tree update by store_directories
        # NULLID is the parent of the first revision
        self._last_hg_nodeid = hgutil.NULLID

        # keeps the last revision tree
        # it is used for differential tree update by store_directories
        self._last_root = HgDirectory()

        # Cache the content hash across revisions to avoid recalculation.
        self._content_hash_cache: hgutil.LRUCacheDict = hgutil.LRUCacheDict(
            content_cache_size,
        )

        # hg node id of the latest snapshot branch heads
        # used to find what are the new revisions since last snapshot
        self._latest_heads: List[bytes] = []
        # hg node ids of all the tags recorded on previous runs
        # Used to compute which tags need to be added, even across incremental runs
        # that might separate a changeset introducing a tag from its target.
        self._saved_tags: Set[bytes] = set()

        self._load_status = "eventful"
        # If set, will override the default value
        self._visit_status = None

    def pre_cleanup(self) -> None:
        """As a first step, will try and check for dangling data to cleanup.
        This should do its best to avoid raising issues.

        """
        clean_dangling_folders(
            self._temp_directory,
            pattern_check=TEMPORARY_DIR_PREFIX_PATTERN,
            log=self.log,
        )

        self.old_environ = os.environ.copy()
        os.environ.clear()
        os.environ.update(get_minimum_env())

    def cleanup(self) -> None:
        """Last step executed by the loader."""
        os.environ.clear()
        os.environ.update(self.old_environ)

        # Don't cleanup if loading from a local directory
        was_remote = self.directory is None
        if was_remote and self._repo_directory and os.path.exists(self._repo_directory):
            self.log.debug(f"Cleanup up repository {self._repo_directory}")
            rmtree(self._repo_directory)

    def prepare_origin_visit(self) -> None:
        """First step executed by the loader to prepare origin and visit
        references. Set/update self.origin, and
        optionally self.origin_url, self.visit_date.

        """
        self.origin = Origin(url=self.origin_url)

    def prepare(self) -> None:
        """Second step executed by the loader to prepare some state needed by
        the loader.

        """
        # Set here to allow multiple calls to load on the same loader instance
        self._latest_heads = []

        latest_snapshot = snapshot_get_latest(self.storage, self.origin_url)
        if latest_snapshot:
            self._set_recorded_state(latest_snapshot)

    def _set_recorded_state(self, latest_snapshot: Snapshot) -> None:
        """
        Looks up the nodeid for all revisions in the snapshot via extid_get_from_target,
        and adds them to `self._latest_heads`.
        Also looks up the currently saved releases ("tags" in Mercurial speak).
        The tags are all listed for easy comparison at the end, while only the latest
        heads are needed for revisions.
        """
        heads = []
        tags = []

        for branch in latest_snapshot.branches.values():
            if branch.target_type == TargetType.REVISION:
                heads.append(branch.target)
            elif branch.target_type == TargetType.RELEASE:
                tags.append(branch.target)

        self._latest_heads.extend(
            extid.extid for extid in self._get_extids_for_targets(heads)
        )
        self._saved_tags.update(
            extid.extid for extid in self._get_extids_for_targets(tags)
        )

    def _get_extids_for_targets(self, targets: List[bytes]) -> List[ExtID]:
        # Get all Mercurial ExtIDs for the targets in the latest snapshot
        extids = [
            extid
            for extid in self.storage.extid_get_from_target(
                identifiers.ObjectType.REVISION, targets
            )
            if extid.extid_type == EXTID_TYPE
        ]

        if extids:
            # Filter out dangling extids, we need to load their target again
            revisions_missing = self.storage.revision_missing(
                [extid.target.object_id for extid in extids]
            )
            extids = [
                extid
                for extid in extids
                if extid.target.object_id not in revisions_missing
            ]
        return extids

    def fetch_data(self) -> bool:
        """Fetch the data from the source the loader is currently loading

        Returns:
            a value that is interpreted as a boolean. If True, fetch_data needs
            to be called again to complete loading.

        """
        if not self.directory:  # no local repository
            self._repo_directory = mkdtemp(
                prefix=TEMPORARY_DIR_PREFIX_PATTERN,
                suffix=f"-{os.getpid()}",
                dir=self._temp_directory,
            )
            self.log.debug(
                f"Cloning {self.origin_url} to {self.directory} "
                f"with timeout {self._clone_timeout} seconds"
            )
            hgutil.clone(self.origin_url, self._repo_directory, self._clone_timeout)
        else:  # existing local repository
            # Allow to load on disk repository without cloning
            # for testing purpose.
            self._repo_directory = self.directory

        self._repo = hgutil.repository(self._repo_directory)

        return False

    def get_hg_revs_to_load(self) -> Union[HgFilteredSet, HgSpanSet]:
        """Return the hg revision numbers to load."""
        assert self._repo is not None
        repo: hgutil.Repository = self._repo
        if self._latest_heads:
            existing_heads = []  # heads that still exist in the repository
            for hg_nodeid in self._latest_heads:
                try:
                    rev = repo[hg_nodeid].rev()
                    existing_heads.append(rev)
                except KeyError:  # the node does not exist anymore
                    pass

            # select revisions that are not ancestors of heads
            # and not the heads themselves
            new_revs = repo.revs("not ::(%ld)", existing_heads)

            if new_revs:
                self.log.info("New revisions found: %d", len(new_revs))
            return new_revs
        else:
            return repo.revs("all()")

    def store_data(self):
        """Store fetched data in the database."""
        revs = self.get_hg_revs_to_load()
        if not revs:
            self._load_status = "uneventful"
            return

        assert self._repo is not None
        repo = self._repo

        ignored_revs: Set[int] = set()
        for rev in revs:
            if rev in ignored_revs:
                continue
            try:
                self.store_revision(repo[rev])
            except CorruptedRevision as e:
                self._visit_status = "partial"
                self.log.warning("Corrupted revision %s", e)
                descendents = repo.revs("(%ld)::", [rev])
                ignored_revs.update(descendents)

        if len(ignored_revs) == len(revs):
            # The repository is completely broken, nothing can be loaded
            self._load_status = "uneventful"
            return

        branching_info = hgutil.branching_info(repo, ignored_revs)
        tags_by_name: Dict[bytes, HgNodeId] = repo.tags()

        snapshot_branches: Dict[bytes, SnapshotBranch] = {}

        for tag_name, hg_nodeid in tags_by_name.items():
            if tag_name == b"tip":
                # `tip` is listed in the tags by the Mercurial API but its not a tag
                # defined by the user in `.hgtags`.
                continue
            if hg_nodeid not in self._saved_tags:
                label = b"tags/%s" % tag_name
                target = self.get_revision_id_from_hg_nodeid(hg_nodeid)
                snapshot_branches[label] = SnapshotBranch(
                    target=self.store_release(tag_name, target),
                    target_type=TargetType.RELEASE,
                )

        for branch_name, node_id in branching_info.tips.items():
            name = b"branch-tip/%s" % branch_name
            target = self.get_revision_id_from_hg_nodeid(node_id)
            snapshot_branches[name] = SnapshotBranch(
                target=target, target_type=TargetType.REVISION
            )

        for bookmark_name, node_id in branching_info.bookmarks.items():
            name = b"bookmarks/%s" % bookmark_name
            target = self.get_revision_id_from_hg_nodeid(node_id)
            snapshot_branches[name] = SnapshotBranch(
                target=target, target_type=TargetType.REVISION
            )

        for branch_name, branch_heads in branching_info.open_heads.items():
            for index, head in enumerate(branch_heads):
                name = b"branch-heads/%s/%d" % (branch_name, index)
                target = self.get_revision_id_from_hg_nodeid(head)
                snapshot_branches[name] = SnapshotBranch(
                    target=target, target_type=TargetType.REVISION
                )

        for branch_name, closed_heads in branching_info.closed_heads.items():
            for index, head in enumerate(closed_heads):
                name = b"branch-closed-heads/%s/%d" % (branch_name, index)
                target = self.get_revision_id_from_hg_nodeid(head)
                snapshot_branches[name] = SnapshotBranch(
                    target=target, target_type=TargetType.REVISION
                )

        # If the repo is broken enough or if it has none of the "normal" default
        # mechanisms, we ignore `HEAD`.
        default_branch_alias = branching_info.default_branch_alias
        if default_branch_alias is not None:
            snapshot_branches[b"HEAD"] = SnapshotBranch(
                target=default_branch_alias, target_type=TargetType.ALIAS,
            )
        snapshot = Snapshot(branches=snapshot_branches)
        self.storage.snapshot_add([snapshot])

        self.flush()
        self.loaded_snapshot_id = snapshot.id

    def load_status(self) -> Dict[str, str]:
        """Detailed loading status.

        Defaults to logging an eventful load.

        Returns: a dictionary that is eventually passed back as the task's
          result to the scheduler, allowing tuning of the task recurrence
          mechanism.
        """
        return {
            "status": self._load_status,
        }

    def visit_status(self) -> str:
        """Allow overriding the visit status in case of partial load"""
        if self._visit_status is not None:
            return self._visit_status
        return super().visit_status()

    def get_revision_id_from_hg_nodeid(self, hg_nodeid: HgNodeId) -> Sha1Git:
        """Return the git sha1 of a revision given its hg nodeid.

        Args:
            hg_nodeid: the hg nodeid of the revision.

        Returns:
            the sha1_git of the revision.
        """

        from_cache = self._revision_nodeid_to_sha1git.get(hg_nodeid)
        if from_cache is not None:
            return from_cache
        # The parent was not loaded in this run, get it from storage
        from_storage = self.storage.extid_get_from_extid(EXTID_TYPE, ids=[hg_nodeid])

        msg = "Expected 1 match from storage for hg node %r, got %d"
        assert len(from_storage) == 1, msg % (hg_nodeid.hex(), len(from_storage))
        return from_storage[0].target.object_id

    def get_revision_parents(self, rev_ctx: hgutil.BaseContext) -> Tuple[Sha1Git, ...]:
        """Return the git sha1 of the parent revisions.

        Args:
            hg_nodeid: the hg nodeid of the revision.

        Returns:
            the sha1_git of the parent revisions.
        """
        parents = []
        for parent_ctx in rev_ctx.parents():
            parent_hg_nodeid = parent_ctx.node()
            # nullid is the value of a parent that does not exist
            if parent_hg_nodeid == hgutil.NULLID:
                continue
            revision_id = self.get_revision_id_from_hg_nodeid(parent_hg_nodeid)
            parents.append(revision_id)

        return tuple(parents)

    def store_revision(self, rev_ctx: hgutil.BaseContext) -> None:
        """Store a revision given its hg nodeid.

        Args:
            rev_ctx: the he revision context.

        Returns:
            the sha1_git of the stored revision.
        """
        hg_nodeid = rev_ctx.node()

        root_sha1git = self.store_directories(rev_ctx)

        # `Person.from_fullname` is compatible with mercurial's freeform author
        # as fullname is what is used in revision hash when available.
        author = Person.from_fullname(rev_ctx.user())

        (timestamp, offset) = rev_ctx.date()

        # TimestampWithTimezone.from_dict will change name
        # as it accept more than just dicts
        rev_date = TimestampWithTimezone.from_dict(int(timestamp))

        extra_headers = [
            (b"time_offset_seconds", str(offset).encode(),),
        ]
        for key, value in rev_ctx.extra().items():
            # The default branch is skipped to match
            # the historical implementation
            if key == b"branch" and value == b"default":
                continue

            # transplant_source is converted to match
            # the historical implementation
            if key == b"transplant_source":
                value = hash_to_bytehex(value)
            extra_headers.append((key, value))

        revision = Revision(
            author=author,
            date=rev_date,
            committer=author,
            committer_date=rev_date,
            type=RevisionType.MERCURIAL,
            directory=root_sha1git,
            message=rev_ctx.description(),
            extra_headers=tuple(extra_headers),
            synthetic=False,
            parents=self.get_revision_parents(rev_ctx),
        )

        self._revision_nodeid_to_sha1git[hg_nodeid] = revision.id
        self.storage.revision_add([revision])

        # Save the mapping from SWHID to hg id
        revision_swhid = identifiers.CoreSWHID(
            object_type=identifiers.ObjectType.REVISION, object_id=revision.id,
        )
        self.storage.extid_add(
            [ExtID(extid_type=EXTID_TYPE, extid=hg_nodeid, target=revision_swhid)]
        )

    def store_release(self, name: bytes, target: Sha1Git) -> Sha1Git:
        """Store a release given its name and its target.

        A release correspond to a user defined tag in mercurial.
        The mercurial api as a `tip` tag that must be ignored.

        Args:
            name: name of the release.
            target: sha1_git of the target revision.

        Returns:
            the sha1_git of the stored release.
        """
        release = Release(
            name=name,
            target=target,
            target_type=ObjectType.REVISION,
            message=None,
            metadata=None,
            synthetic=False,
            author=Person(name=None, email=None, fullname=b""),
            date=None,
        )

        self.storage.release_add([release])

        return release.id

    def store_content(self, rev_ctx: hgutil.BaseContext, file_path: bytes) -> Content:
        """Store a revision content hg nodeid and file path.

        Content is a mix of file content at a given revision
        and its permissions found in the changeset's manifest.

        Args:
            rev_ctx: the he revision context.
            file_path: the hg path of the content.

        Returns:
            the sha1_git of the top level directory.
        """
        hg_nodeid = rev_ctx.node()
        file_ctx = rev_ctx[file_path]

        try:
            file_nodeid = file_ctx.filenode()
        except hgutil.LookupError:
            # TODO
            # Raising CorruptedRevision avoid crashing the whole loading
            # but can lead to a lot of missing revisions.
            # SkippedContent could be used but need actual content to calculate its id.
            # Maybe the hg_nodeid can be used instead.
            # Another option could be to just ignore the missing content.
            # This point is left to future commits.
            # Check for other uses to apply the same logic there.
            raise CorruptedRevision(hg_nodeid)

        perms = FLAG_PERMS[file_ctx.flags()]

        # Key is file_nodeid + perms because permissions does not participate
        # in content hash in hg while it is the case in swh.
        cache_key = (file_nodeid, perms)

        sha1_git = self._content_hash_cache.get(cache_key)
        if sha1_git is None:
            try:
                data = file_ctx.data()
            except hgutil.error.RevlogError:
                # TODO
                # See above use of `CorruptedRevision`
                raise CorruptedRevision(hg_nodeid)

            content = ModelContent.from_data(data)

            self.storage.content_add([content])

            sha1_git = content.sha1_git
            self._content_hash_cache[cache_key] = sha1_git

        # Here we make sure to return only necessary data.
        return Content({"sha1_git": sha1_git, "perms": perms})

    def store_directories(self, rev_ctx: hgutil.BaseContext) -> Sha1Git:
        """Store a revision directories given its hg nodeid.

        Mercurial as no directory as in git. A Git like tree must be build
        from file paths to obtain each directory hash.

        Args:
            rev_ctx: the he revision context.

        Returns:
            the sha1_git of the top level directory.
        """
        repo: hgutil.Repository = self._repo  # mypy can't infer that repo is not None
        prev_ctx = repo[self._last_hg_nodeid]

        # TODO maybe do diff on parents
        try:
            status = prev_ctx.status(rev_ctx)
        except hgutil.error.LookupError:
            raise CorruptedRevision(rev_ctx.node())

        for file_path in status.removed:
            try:
                del self._last_root[file_path]
            except KeyError:
                raise CorruptedRevision(rev_ctx.node())

        for file_path in status.added:
            content = self.store_content(rev_ctx, file_path)
            self._last_root[file_path] = content

        for file_path in status.modified:
            content = self.store_content(rev_ctx, file_path)
            self._last_root[file_path] = content

        self._last_hg_nodeid = rev_ctx.node()

        directories: Deque[Directory] = deque([self._last_root])
        while directories:
            directory = directories.pop()
            self.storage.directory_add([directory.to_model()])
            directories.extend(
                [item for item in directory.values() if isinstance(item, Directory)]
            )

        return self._last_root.hash


class HgArchiveLoaderFromDisk(HgLoaderFromDisk):
    """Mercurial loader for repository wrapped within tarballs."""

    def __init__(
        self,
        storage: StorageInterface,
        url: str,
        visit_date: Optional[datetime] = None,
        archive_path: str = None,
        temp_directory: str = "/tmp",
        max_content_size: Optional[int] = None,
    ):
        super().__init__(
            storage=storage,
            url=url,
            visit_date=visit_date,
            logging_class="swh.loader.mercurial.ArchiveLoaderFromDisk",
            temp_directory=temp_directory,
            max_content_size=max_content_size,
        )
        self.archive_extract_temp_dir = None
        self.archive_path = archive_path

    def prepare(self):
        """Extract the archive instead of cloning."""
        self.archive_extract_temp_dir = tmp_extract(
            archive=self.archive_path,
            dir=self._temp_directory,
            prefix=TEMPORARY_DIR_PREFIX_PATTERN,
            suffix=f".dump-{os.getpid()}",
            log=self.log,
            source=self.origin_url,
        )

        repo_name = os.listdir(self.temp_dir)[0]
        self.directory = os.path.join(self.archive_extract_temp_dir, repo_name)
        super().prepare()


# Allow direct usage of the loader from the command line with
# `python -m swh.loader.mercurial.from_disk $ORIGIN_URL`
if __name__ == "__main__":
    import logging

    import click

    logging.basicConfig(
        level=logging.DEBUG, format="%(asctime)s %(process)d %(message)s"
    )

    @click.command()
    @click.option("--origin-url", help="origin url")
    @click.option("--hg-directory", help="Path to mercurial repository to load")
    @click.option("--visit-date", default=None, help="Visit date")
    def main(origin_url, hg_directory, visit_date):
        from swh.storage import get_storage

        storage = get_storage(cls="memory")
        return HgLoaderFromDisk(
            storage,
            origin_url,
            directory=hg_directory,
            visit_date=parse_visit_date(visit_date),
        ).load()

    main()



# Copyright (C) 2017-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Optional

from celery import shared_task

from swh.loader.mercurial.utils import parse_visit_date

from .from_disk import HgArchiveLoaderFromDisk, HgLoaderFromDisk


@shared_task(name=__name__ + ".LoadMercurial")
def load_hg(
    *, url: str, directory: Optional[str] = None, visit_date: Optional[str] = None
):
    """Mercurial repository loading

    Import a mercurial tarball into swh.

    Args: see :func:`HgLoaderFromDisk.load`.

    """

    loader = HgLoaderFromDisk.from_configfile(
        url=url, directory=directory, visit_date=parse_visit_date(visit_date)
    )
    return loader.load()


@shared_task(name=__name__ + ".LoadArchiveMercurial")
def load_hg_from_archive(
    *, url: str, archive_path: Optional[str] = None, visit_date: Optional[str] = None
):
    """Import a mercurial tarball into swh.

    Args: see :func:`HgArchiveLoaderFromDisk.load`.
    """
    loader = HgArchiveLoaderFromDisk.from_configfile(
        url=url, archive_path=archive_path, visit_date=parse_visit_date(visit_date)
    )
    return loader.load()



# Copyright (C) 2018-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from itertools import chain
import logging

import click
from deprecated import deprecated

from swh.loader.mercurial.utils import parse_visit_date

LOGLEVELS = list(
    chain.from_iterable(
        (logging._levelToName[lvl], logging._levelToName[lvl].lower())
        for lvl in sorted(logging._levelToName.keys())
    )
)


@click.command()
@click.argument("origin-url")
@click.option(
    "--hg-directory",
    "-d",
    help=(
        "Path to the hg (local) directory to load from. "
        "If unset, the hg repo will be cloned from the "
        "given (origin) url."
    ),
)
@click.option("--hg-archive", "-a", help=("Path to the hg archive file to load from."))
@click.option("--visit-date", "-D", help="Visit date (defaults to now).")
@click.option("--log-level", "-l", type=click.Choice(LOGLEVELS), help="Log level.")
@deprecated(
    version="0.4.0", reason="Use `swh loader run mercurial|mercurial_from_disk` instead"
)
def main(
    origin_url, hg_directory=None, hg_archive=None, visit_date=None, log_level=None
):
    from swh.storage import get_storage

    logging.basicConfig(
        level=(log_level or "DEBUG").upper(),
        format="%(asctime)s %(process)d %(message)s",
    )

    visit_date = parse_visit_date(visit_date or "now")

    kwargs = {"visit_date": visit_date, "url": origin_url}
    if hg_archive:
        from .loader import HgArchiveBundle20Loader as HgLoader

        kwargs["archive_path"] = hg_archive
    else:
        from .loader import HgBundle20Loader as HgLoader

        kwargs["directory"] = hg_directory

    storage = get_storage(cls="memory")
    return HgLoader(storage, **kwargs).load()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
"""
This build script purpose is to create example repositories from bash scripts
and extract assertion data from them into json files.

Advantages:

    - the bash script documents the repository creation
    - automating creation allow easy repository update
    - automation extraction allow easier update of assertion data

Create a new repository
-----------------------

First, create a bash script `myscript.sh` which will serve to create, update
and document the repository.

Here is a minimal working example:

    #!/usr/bin/env bash

    # Setup bash in strict mode
    set -euo pipefail

    # Allow direct call to call the script: `./myscript.sh repository-name`
    if [ -n "$1" ]; then
        HG_REPO="$1"
    fi

    # Prepare the repository
    hg init "$HG_REPO"
    cd "$HG_REPO"
    cat > .hg/hgrc << EOL
    [ui]
    username = Full Name<full.name@domain.tld>
    EOL

    # Populate the repository
    touch README.md
    hg add README.md
    hg commit -m "Add README"

Then build the repository and the associated json file which containing
the repository objects identities:

    ./build.py json myscript.sh

You should now have a `myscript.tgz` containing the repository
and a `myscript.json` with the repository objects identities.

Update a repository
-------------------

When there is a build script file:

    Update the source `repository.sh` and run `./build.py json repository.sh`.
    New `tgz` and `json` files will be produced.
    The old files will be renamed.

When There is no build script file:

    Maybe consider not updating the repository and add a new one instead.

    Otherwise, uncompress the `repository.tgz`, manually update the `repository`
    and run `./build.py json repository`.
    New `tgz` and `json` files will be produced.
    The old files will be renamed.

Extract json from existing repository
-------------------------------------

For existing repository without build scripts it is possible to extract
the corresponding json file by running `./build.py json repository.tgz`
"""

from datetime import datetime
import json
from pathlib import Path
import shutil
import subprocess

import click


def abort(message):
    """Abort the script with a message."""
    click.echo(message, err=True)
    click.get_current_context().abort()


def backup(path: Path):
    """Rename an existing path."""
    click.echo(f"Creating backup of {path}")
    now = datetime.now()
    backup_path = path.with_suffix(f"{path.suffix}.bak.{now:%Y%m%d%H%M%S}")
    path.rename(backup_path)
    click.echo(f"Backup created: {str(backup_path)!r}")


def _build_repository(script: str) -> Path:
    """Build a repository from a bash script."""
    script_path = Path(script).absolute()

    if not script_path.exists():
        abort(f"Path {script_path!r} does not exists.")

    if script_path.suffix != ".sh":
        abort(f"Wrong suffix: {script_path.suffix!r}. Expected: '.sh'")

    repository_path = script_path.with_suffix("")

    if repository_path.exists():
        backup(repository_path)

    click.echo(f"Running build script: {str(script_path)!r}")
    subprocess.run(
        ["bash", "-eo", "pipefail", script_path, repository_path.stem],
        env={"HG_REPO": str(repository_path), "HGPLAIN": "", "HGRCPATH": ""},
    )

    return repository_path


def _build_json(source: str) -> Path:
    if source.endswith(".tgz"):
        archive_path = Path(source).absolute()
        repository_path = archive_path.with_suffix("")

        if repository_path.exists():
            backup(repository_path)

        subprocess.call(["tar", "-xf", archive_path], cwd=archive_path.parent)
    elif source.endswith(".sh"):
        repository_path = _build_repository(source)
    else:
        repository_path = Path(source).absolute()

    click.echo(f"Extracting object identities: {str(repository_path)!r}")
    output = subprocess.check_output(["swh-hg-identify", "all"], cwd=repository_path)
    lines = output.decode().splitlines()

    directory_swhids = []
    revision_swhids = []
    release_swhids = []

    for line in lines:
        uri, _ = line.split("\t")
        _, _, swhid_type, swhid = uri.split(":")
        if swhid_type == "dir":
            directory_swhids.append(swhid)
        elif swhid_type == "rev":
            revision_swhids.append(swhid)
        elif swhid_type == "rel":
            release_swhids.append(swhid)
        elif swhid_type == "snp":
            snapshot_swhid = swhid
        else:
            abort(f"{line!r} unknown type {swhid_type!r}")

    json_path = repository_path.with_suffix(".json")

    if json_path.exists():
        backup(json_path)

    click.echo(f"Creating object identities file: {str(json_path)!r}")
    json_path.write_text(
        json.dumps(
            {
                "directories": directory_swhids,
                "revisions": revision_swhids,
                "releases": release_swhids,
                "snapshot": snapshot_swhid,
            }
        )
    )

    return json_path


@click.group(help=__doc__.split("\n\n")[0])
def main():
    pass


@main.command()
def man():
    """Display script's manual."""
    click.echo(__doc__)


@main.command("repository")
@click.argument("script")
def build_repository(script: str):
    """Build a repository.

    SCRIPT must be is a bash script with a `.sh` suffix

    The generated repository will have the same path minor the `.sh` suffix.

    The script will be passed repository name as the `HG_REPO` environment variable.
    """
    _build_repository(script)


@main.command("json")
@click.argument("source")
def build_json(source: str):
    """Build a json file of object identities.

    SOURCE can be a script as required by the `repository` command
    (see repository --help), a repository archive, or an existing repository.

    The produced file will have the source path with the `.json` suffix.
    """
    _build_json(source)


@main.command("archive")
@click.option(
    "--clean", "-c", default=False, is_flag=True, help="Remove created artifacts",
)
@click.argument("source")
def build_archive(source: str, clean: bool = False):
    """Build a repository archive and is associated json file.

    SOURCE can be a script as required by the `repository` command
    (see repository --help), or an existing repository.

    The produced archive will have the source path with the `.tgz` suffix.
    The produced json file will have the source path with the `.json` suffix.
    """
    if source.endswith(".sh"):
        repository_path = _build_repository(source)
    else:
        repository_path = Path(source).absolute()
        if not (repository_path / ".hg").exists():
            abort(f"{str(repository_path)!r} is not a Mercurial repository")

    json_path = _build_json(str(repository_path))

    archive_path = repository_path.with_suffix(".tgz")
    if archive_path.exists():
        backup(archive_path)

    subprocess.call(
        [
            "tar",
            "-cf",
            archive_path.relative_to(archive_path.parent),
            repository_path.relative_to(archive_path.parent),
            json_path.relative_to(archive_path.parent),
        ],
        cwd=archive_path.parent,
    )

    if clean:
        shutil.rmtree(repository_path)
        json_path.unlink()


if __name__ == "__main__":
    main()



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""This document contains various helper classes used in converting Mercurial
bundle files into SWH Contents, Directories, etc.
"""

import binascii
from collections import OrderedDict
import copy
import os
import pickle
import sqlite3
import sys
import zlib

from sqlitedict import SqliteDict

from swh.model import identifiers

OS_PATH_SEP = os.path.sep.encode("utf-8")


def _encode(obj):
    return sqlite3.Binary(zlib.compress(pickle.dumps(obj, pickle.HIGHEST_PROTOCOL)))


def _decode(obj):
    return pickle.loads(zlib.decompress(bytes(obj)))


class SimpleBlob:
    """Stores basic metadata of a blob object.when constructing deep trees from
    commit file manifests.

    args:
        file_hash: unique hash of the file contents
        is_symlink: (bool) is this file a symlink?
        file_perms: (string) 3 digit permission code as a string or bytestring,
                    e.g. '755' or b'755'
    """

    kind = "file"

    def __init__(self, file_hash, is_symlink, file_perms):
        self.hash = file_hash
        self.perms = 0o100000 + int(file_perms, 8)
        if is_symlink:
            self.perms += 0o020000

    def __str__(self):
        return "SimpleBlob: " + str(self.hash) + " -- " + str(self.perms)

    def __eq__(self, other):
        return (self.perms == other.perms) and (self.hash == other.hash)

    def size(self):
        """Return the size in byte."""
        return sys.getsizeof(self) + sys.getsizeof(self.__dict__)


class SimpleTree(dict):
    """ Stores data for a nested directory object. Uses shallow cloning to stay
    compact after forking and change monitoring for efficient re-hashing.
    """

    kind = "dir"
    perms = 0o040000

    def __init__(self):
        self.hash = None
        self._size = None

    def __eq__(self, other):
        return (self.hash == other.hash) and (self.items() == other.items())

    def _new_tree_node(self, path):
        """Deeply nests SimpleTrees according to a given subdirectory path and
        returns a reference to the deepest one.

        args:
            path: bytestring containing a relative path from self to a deep
                  subdirectory. e.g. b'foodir/bardir/bazdir'

        returns:
            the new node
        """
        node = self
        for d in path.split(OS_PATH_SEP):
            if node.get(d):
                if node[d].hash is not None:
                    node[d] = copy.copy(node[d])
                    node[d].hash = None
                    node[d]._size = None
            else:
                node[d] = SimpleTree()
            node = node[d]
        return node

    def remove_tree_node_for_path(self, path):
        """Deletes a SimpleBlob or SimpleTree from inside nested SimpleTrees
        according to the given relative file path, and then recursively removes
        any newly depopulated SimpleTrees. It keeps the old history by doing a
        shallow clone before any change.

        args:
            path: bytestring containing a relative path from self to a nested
                  file or directory. e.g. b'foodir/bardir/bazdir/quxfile.txt'

        returns:
            the new root node
        """
        node = self
        if node.hash is not None:
            node = copy.copy(node)
            node.hash = None
            node._size = None
        first, sep, rest = path.partition(OS_PATH_SEP)
        if rest:
            node[first] = node[first].remove_tree_node_for_path(rest)
            if len(node[first]) == 0:
                del node[first]
        else:
            del node[first]

        return node

    def add_blob(self, file_path, file_hash, is_symlink, file_perms):
        """Shallow clones the root node and then deeply nests a SimpleBlob
        inside nested SimpleTrees according to the given file path, shallow
        cloning all all intermediate nodes and marking them as changed and
        in need of new hashes.

        args:
            file_path: bytestring containing the relative path from self to a
                       nested file
            file_hash: primary identifying hash computed from the blob contents
            is_symlink: True/False whether this item is a symbolic link
            file_perms: int or string representation of file permissions

        returns:
            the new root node
        """
        root = self
        if root.hash is not None:
            root = copy.copy(root)
            root.hash = None
            root._size = None
        node = root
        fdir, fbase = os.path.split(file_path)
        if fdir:
            node = root._new_tree_node(fdir)
        node[fbase] = SimpleBlob(file_hash, is_symlink, file_perms)
        return root

    def yield_swh_directories(self):
        """Converts nested SimpleTrees into a stream of SWH Directories.

        yields:
            an SWH Directory for every node in the tree
        """
        for k, v in sorted(self.items()):
            if isinstance(v, SimpleTree):
                yield from v.yield_swh_directories()

        yield {
            "id": self.hash,
            "entries": [
                {"name": k, "perms": v.perms, "type": v.kind, "target": v.hash}
                for k, v in sorted(self.items())
            ],
        }

    def hash_changed(self, new_dirs=None):
        """Computes and sets primary identifier hashes for unhashed subtrees.

        args:
            new_dirs (optional): an empty list to be populated with the SWH
                                 Directories for all of the new (not previously
                                 hashed) nodes

        returns:
            the top level hash of the whole tree
        """
        if self.hash is None:
            directory = {
                "entries": [
                    {
                        "name": k,
                        "perms": v.perms,
                        "type": v.kind,
                        "target": (
                            v.hash if v.hash is not None else v.hash_changed(new_dirs)
                        ),
                    }
                    for k, v in sorted(self.items())
                ]
            }

            self.hash = binascii.unhexlify(identifiers.directory_identifier(directory))
            directory["id"] = self.hash
            if new_dirs is not None:
                new_dirs.append(directory)

        return self.hash

    def flatten(self, _curpath=None, _files=None):
        """Converts nested sub-SimpleTrees and SimpleBlobs into a list of
        file paths. Useful for counting the number of files in a manifest.

        returns:
            a flat list of all of the contained file paths
        """
        _curpath = _curpath or b""
        _files = _files or {}
        for k, v in sorted(self.items()):
            p = os.path.join(_curpath, k)
            if isinstance(v, SimpleBlob):
                _files[p] = (v.hash, v.perms)
            else:
                v.flatten(p, _files)
        return _files

    def size(self):
        """Return the (approximate?) memory utilization in bytes of the nested
        structure.
        """
        if self._size is None:
            self._size = (
                sys.getsizeof(self)
                + sys.getsizeof(self.__dict__)
                + sum([sys.getsizeof(k) + v.size() for k, v in self.items()])
            )
        return self._size


class SelectiveCache(OrderedDict):
    """Special cache for storing past data upon which new data is known to be
    dependent. Optional hinting of how many instances of which keys will be
    needed down the line makes utilization more efficient. And, because the
    distance between related data can be arbitrarily long and the data
    fragments can be arbitrarily large, a disk-based secondary storage is used
    if the primary RAM-based storage area is filled to the designated capacity.

    Storage is occupied in three phases:

        1) The most recent key/value pair is always held, regardless of other
        factors, until the next entry replaces it.

        2) Stored key/value pairs are pushed into a randomly accessible
        expanding buffer in memory with a stored size function, maximum size
        value, and special hinting about which keys to store for how long
        optionally declared at instantiation.

        3) The in-memory buffer pickles into a randomly accessible disk-backed
        secondary buffer when it becomes full.

    Occupied space is calculated by default as whatever the len() function
    returns on the values being stored. This can be changed by passing in a new
    size_function at instantiation.

    The cache_hints parameter is a dict of key/int pairs recording how many
    subsequent fetches that particular key's value should stay in storage for
    before being erased. If you provide a set of hints and then try to store a
    key that is not in that set of hints, the cache will store it only while it
    is the most recent entry, and will bypass storage phases 2 and 3.
    """

    DEFAULT_SIZE = 800 * 1024 * 1024  # bytes or whatever

    def __init__(
        self, max_size=None, cache_hints=None, size_function=None, filename=None
    ):
        """
        args:
            max_size: integer value indicating the maximum size of the part
                of storage held in memory
            cache_hints: dict of key/int pairs as described in the class
                description
            size_function: callback function that accepts one parameter and
                returns one int, which should probably be the calculated
                size of the parameter
        """
        self._max_size = max_size or SelectiveCache.DEFAULT_SIZE
        self._disk = None
        if size_function is None:
            self._size_function = sys.getsizeof
        else:
            self._size_function = size_function
        self._latest = None
        self._cache_size = 0
        self._cache_hints = copy.copy(cache_hints) or None
        self.filename = filename

    def store(self, key, data):
        """Primary method for putting data into the cache.

        args:
            key: any hashable value
            data: any python object (preferably one that is measurable)

        """
        self._latest = (key, data)

        if (self._cache_hints is not None) and (key not in self._cache_hints):
            return

        # cache the completed data...
        self._cache_size += self._size_function(data) + 53

        # ...but limit memory expenditure for the cache by offloading to disk
        should_commit = False
        while self._cache_size > self._max_size and len(self) > 0:
            should_commit = True
            k, v = self.popitem(last=False)
            self._cache_size -= self._size_function(v) - 53
            self._diskstore(k, v)

        if should_commit:
            self._disk.commit(blocking=False)

        self[key] = data

    def _diskstore(self, key, value):
        if self._disk is None:
            self._disk = SqliteDict(
                autocommit=False,
                journal_mode="OFF",
                filename=self.filename,
                tablename="swh",
                encode=_encode,
                decode=_decode,
            )
            self._disk.in_temp = True  # necessary to force the disk clean up
        self._disk[key] = value

    def has(self, key):
        """Tests whether the data for the provided key is being stored.

        args:
            key: the key of the data whose storage membership property you wish
                 to discover

        returns:
            True or False
        """
        return (
            (self._latest and self._latest[0] == key)
            or (key in self)
            or (self._disk and (key in self._disk))
        )

    def fetch(self, key):
        """Pulls a value out of storage and decrements the hint counter for the
            given key.

        args:
            key: the key of the data that you want to retrieve

        returns:
            the retrieved value or None
        """
        retval = None
        if self._latest and self._latest[0] == key:
            retval = self._latest[1]
        if retval is None:
            retval = self.get(key)
            if (retval is None) and self._disk:
                self._disk.commit(blocking=False)
                retval = self._disk.get(key) or None
            self.dereference(key)
        return retval

    def dereference(self, key):
        """Remove one instance of expected future retrieval of the data for the
        given key. This is called automatically by fetch requests that aren't
        satisfied by phase 1 of storage.

        args:
            the key of the data for which the future retrievals hint is to be
            decremented
        """
        newref = self._cache_hints and self._cache_hints.get(key)
        if newref:
            newref -= 1
            if newref == 0:
                del self._cache_hints[key]
                if key in self:
                    item = self[key]
                    self._cache_size -= self._size_function(item)
                    del self[key]
                else:
                    if self._disk:
                        del self._disk[key]
            else:
                self._cache_hints[key] = newref

    def keys(self):
        yield from self.keys()
        if self._disk:
            yield from self._disk.keys()

    def values(self):
        yield from self.values()
        if self._disk:
            yield from self._disk.values()

    def items(self):
        yield from self.items()
        if self._disk:
            yield from self._disk.items()
